/*
    Stuphead: (C) 2000,2001 Grigroy Bakunov, Sergey Pinaev
 */
/*
 * Sylpheed -- a GTK+ based, lightweight, and fast e-mail client
 * Copyright (C) 1999,2000 Hiroyuki Yamamoto
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#ifdef GNOME_SUPPORT
#  include <gnome.h>
#endif

#include "intl.h"
#include "alertpanel.h"
#include "manage_window.h"
#include "utils.h"
#include "gtkutils.h"

#ifndef GNOME_SUPPORT
#define MESSAGE_FONT	"-*-helvetica-medium-r-normal--14-*-*-*-*-*-*-*," \
			"-*-*-medium-r-normal--14-*-*-*-*-*-*-*,*"
#define TITLE_FONT	"-*-helvetica-medium-r-normal--17-*-*-*-*-*-*-*," \
			"-*-*-medium-r-normal--16-*-*-*-*-*-*-*,*"
#define ALERT_PANEL_WIDTH	380
#define TITLE_HEIGHT		72
#define MESSAGE_HEIGHT		62

static gboolean alertpanel_is_open = FALSE;
static AlertValue value;
static GtkWidget *dialog;

static void alertpanel_show		(void);
static void alertpanel_create		(const gchar	*title,
					 const gchar	*message,
					 const gchar	*button1_label,
					 const gchar	*button2_label,
					 const gchar	*button3_label);
static void alertpanel_button_clicked	(GtkWidget	*widget,
					 gpointer	 data);
static void alertpanel_close		(GtkWidget	*widget,
					 GdkEventAny	*event,
					 gpointer	 data);
#else /* GNOME_SUPPORT */
void gnome_msg_box(const gchar *title, 
		 const gchar *message,
		 const gchar *type);
#endif /* GNOME_SUPPORT */

AlertValue alertpanel(const gchar *title,
		      const gchar *message,
		      const gchar *button1_label,
		      const gchar *button2_label,
		      const gchar *button3_label)
{
#ifndef GNOME_SUPPORT
	if (alertpanel_is_open)
		return -1;
	else
		alertpanel_is_open = TRUE;

	alertpanel_create(title, message, button1_label, button2_label,
			  button3_label);
	alertpanel_show();

	debug_print("return value = %d\n", value);
	return value;
#else /* GNOME_SUPPORT */
	GtkWidget *dialog;
	GList *button;
	gint defbutt;
	gint value;
	gchar *lab1, *lab2, *lab3;

	lab1 = button1_label;
	lab2 = button2_label;
	lab3 = button3_label;

	defbutt = 0;
	if (lab1) {
		if (*lab1 == '+') {
			lab1++; 
		}
	} else
		lab1 = GNOME_STOCK_BUTTON_OK;
		
	if (lab2 && *lab2 == '+') {
		lab2++;
		defbutt = 1;
	};
	if (lab3 && *lab3 == '+') {
		lab3++;
		defbutt = 2;
	};
	debug_print("default button = %d\n", defbutt);

	dialog = gnome_message_box_new(message, GNOME_MESSAGE_BOX_QUESTION,
				       lab1, lab2, lab3, NULL);

	/* set default buttons */
	button = g_list_first(GNOME_DIALOG(dialog)->buttons);
	while (defbutt && button) {
		button = g_list_next(button);
		defbutt--;
	};
	gtk_widget_grab_default(button->data);
	gtk_widget_grab_focus(button->data);

	if (title) gtk_window_set_title(GTK_WINDOW(dialog), title);

	value = gnome_dialog_run_and_close(GNOME_DIALOG(dialog));
	debug_print("alert panel return value = %d\n", value);

	return value;
#endif /* GNOME_SUPPORT */
}

void alertpanel_message(const gchar *title, const gchar *message)
{
#ifndef GNOME_SUPPORT
	if (alertpanel_is_open)
		return;
	else
		alertpanel_is_open = TRUE;

	alertpanel_create(title, message, NULL, NULL, NULL);
	alertpanel_show();
#else /* GNOME_SUPPORT */
	gnome_msg_box(title, message, GNOME_MESSAGE_BOX_GENERIC);
#endif /* GNOME_SUPPORT */
}

void alertpanel_notice(const gchar *format, ...)
{
	va_list args;
	gchar buf[256];

	va_start(args, format);
	g_vsnprintf(buf, sizeof(buf), format, args);
	va_end(args);
	strretchomp(buf);

#ifndef GNOME_SUPPORT
	alertpanel_message(_("Notice"), buf);
#else /* GNOME_SUPPORT */
	gnome_msg_box(_("Notice"), buf, GNOME_MESSAGE_BOX_INFO);
#endif /* GNOME_SUPPORT */
}

void alertpanel_warning(const gchar *format, ...)
{
	va_list args;
	gchar buf[256];

	va_start(args, format);
	g_vsnprintf(buf, sizeof(buf), format, args);
	va_end(args);
	strretchomp(buf);

#ifndef GNOME_SUPPORT
	alertpanel_message(_("Warning"), buf);
#else /* GNOME_SUPPORT */
	gnome_msg_box(NULL, buf, GNOME_MESSAGE_BOX_WARNING);
#endif /* GNOME_SUPPORT */
}

void alertpanel_error(const gchar *format, ...)
{
	va_list args;
	gchar buf[256];

	va_start(args, format);
	g_vsnprintf(buf, sizeof(buf), format, args);
	va_end(args);
	strretchomp(buf);

#ifndef GNOME_SUPPORT
	alertpanel_message(_("Error"), buf);
#else /* GNOME_SUPPORT */
	gnome_msg_box(NULL, buf, GNOME_MESSAGE_BOX_ERROR);
#endif /* GNOME_SUPPORT */
}

#ifndef GNOME_SUPPORT
static void alertpanel_show(void)
{
	gtk_window_set_modal(GTK_WINDOW(dialog), TRUE);
	manage_window_set_transient(GTK_WINDOW(dialog));
	value = G_ALERTWAIT;

	while (value == G_ALERTWAIT)
		gtk_main_iteration();

	gtk_widget_destroy(dialog);
	GTK_EVENTS_FLUSH();

	alertpanel_is_open = FALSE;
}

static void alertpanel_create(const gchar *title,
			      const gchar *message,
			      const gchar *button1_label,
			      const gchar *button2_label,
			      const gchar *button3_label)
{
	static GdkFont *messagefont;
	GtkStyle *style;
	GtkWidget *label;
	GtkWidget *hbox;
	GtkWidget *vbox;
	GtkWidget *confirm_area;
	GtkWidget *button1;
	GtkWidget *button2;
	GtkWidget *button3;
	const gchar *label2;
	const gchar *label3;

	GtkWidget *frame;

	debug_print(_("Creating alert panel dialog...\n"));

	dialog = gtk_dialog_new();
	gtk_window_set_title(GTK_WINDOW(dialog), title);
	gtk_window_set_policy(GTK_WINDOW(dialog), FALSE, FALSE, FALSE);
//	gtk_widget_set_usize(dialog, ALERT_PANEL_WIDTH, -1);
	gtk_container_set_border_width
		(GTK_CONTAINER(GTK_DIALOG(dialog)->action_area), 5);
	gtk_window_position(GTK_WINDOW(dialog), GTK_WIN_POS_CENTER);
        gtk_container_set_border_width(GTK_CONTAINER(dialog), 5);
	gtk_signal_connect(GTK_OBJECT(dialog), "delete_event",
			   GTK_SIGNAL_FUNC(alertpanel_close),
			   (gpointer)G_ALERTOTHER);
	gtk_signal_connect(GTK_OBJECT(dialog), "key_press_event",
			   GTK_SIGNAL_FUNC(alertpanel_close),
			   (gpointer)G_ALERTOTHER);
	gtk_widget_realize(dialog);

        frame=gtk_frame_new(title);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox),
			   frame, TRUE, TRUE, 5);

	hbox = gtk_hbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(frame), hbox);

	/* message label */
	label = gtk_label_new(message);
	gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 10);
	style = gtk_style_copy(gtk_widget_get_style(label));
	if (!messagefont)
		messagefont = gdk_fontset_load(MESSAGE_FONT);
	if (messagefont)
		style->font = messagefont;
	gtk_widget_set_style(label, style);

	/* for message and button(s) */
	vbox = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->action_area),
			  vbox);

	/* for button(s) */
	if (!button1_label)
		button1_label = _("OK");
	label2 = button2_label;
	label3 = button3_label;
	if (label2 && *label2 == '+') label2++;
	if (label3 && *label3 == '+') label3++;

	gtkut_button_set_create(&confirm_area,
				&button1, button1_label,
				button2_label ? &button2 : NULL, label2,
				button3_label ? &button3 : NULL, label3);

	gtk_box_pack_end(GTK_BOX(vbox), confirm_area, FALSE, FALSE, 0);
	gtk_widget_grab_default(button1);
	gtk_widget_grab_focus(button1);
	if (button2_label && *button2_label == '+') {
		gtk_widget_grab_default(button2);
		gtk_widget_grab_focus(button2);
	}
	if (button3_label && *button3_label == '+') {
		gtk_widget_grab_default(button3);
		gtk_widget_grab_focus(button3);
	}

	gtk_signal_connect(GTK_OBJECT(button1), "clicked",
			   GTK_SIGNAL_FUNC(alertpanel_button_clicked),
			   (gpointer)G_ALERTDEFAULT);
	if (button2_label)
		gtk_signal_connect(GTK_OBJECT(button2), "clicked",
				   GTK_SIGNAL_FUNC(alertpanel_button_clicked),
				   (gpointer)G_ALERTALTERNATE);
	if (button3_label)
		gtk_signal_connect(GTK_OBJECT(button3), "clicked",
				   GTK_SIGNAL_FUNC(alertpanel_button_clicked),
				   (gpointer)G_ALERTOTHER);

	gtk_widget_show_all(dialog);
}

static void alertpanel_button_clicked(GtkWidget *widget, gpointer data)
{
	value = (AlertValue)data;
}

static void alertpanel_close(GtkWidget *widget, GdkEventAny *event,
			     gpointer data)
{
	if (event->type == GDK_KEY_PRESS)
		if (((GdkEventKey *)event)->keyval != GDK_Escape)
			return;

	value = (AlertValue)data;
}

#else /* GNOME_SUPPORT */

void gnome_msg_box(const gchar *title,
		 const gchar *message,
		 const gchar *type)
{
	GtkWidget *dialog;

	dialog = gnome_message_box_new(message, type,
				       GNOME_STOCK_BUTTON_OK,
				       NULL);
	if (title) gtk_window_set_title(GTK_WINDOW(dialog), title);
	gnome_dialog_run_and_close(GNOME_DIALOG(dialog));		
}

#endif /* GNOME_SUPPORT */
