/*
    Stuphead: (C) 2000,2001 Grigroy Bakunov, Sergey Pinaev
 */
/*
 * Sylpheed -- a GTK+ based, lightweight, and fast e-mail client
 * Copyright (C) 1999,2000 Hiroyuki Yamamoto
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <glib.h>
#include <gtk/gtkmain.h>
#include <gtk/gtkwindow.h>
#include <gtk/gtksignal.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <signal.h>
#include <errno.h>

#include "main.h"
#include "inc.h"
#include "mainwindow.h"
#include "folderview.h"
#include "prefs_common.h"
#include "prefs_account.h"
#include "account.h"
#include "procmsg.h"
#include "socket.h"
#include "pop.h"
#include "recv.h"
#include "mbox.h"
#include "utils.h"
#include "gtkutils.h"
#include "statusbar.h"
#include "manage_window.h"
#include "progressdialog.h"
#include "inputdialog.h"
#include "alertpanel.h"
#include "filter.h"
#include "automaton.h"

#define MSGBUFSIZE	8192
#define TMPMSG 		".tmpmsg"

static void pop3_automaton_create	(PrefsAccount	*ac_prefs,
					 MainWindow	*mainwin,
					 IncState	*retval);
static void pop3_automaton_destroy	(Automaton	*atm);
static gint pop3_terminate		(gint		 source,
					 Automaton	*atm);
static GHashTable *inc_get_uidl_table	(PrefsAccount	*ac_prefs);
static void inc_write_uidl_list		(Pop3State	*state);

#if USE_THREADS
static void pop3_receive_threaded	(SockInfo	*sockinfo);
static gint connection_check_cb		(Automaton	*atm);
#endif

static gint inc_account_mail		(PrefsAccount	*ac_prefs,
					 gpointer	 data);
static gint inc_account_mail_by_timer	(PrefsAccount *ac_prefs,
					 gpointer data);
static ProgressDialog *inc_progress_create
					(gpointer	 data);
static void inc_cancel			(GtkWidget	*widget,
                                         gpointer	 data);
static gint inc_spool			(void);
static gint get_spool			(const gchar	*dest,
					 const gchar	*mbox);
static void inc_update(MainWindow *mainwin)
{
    gint msgnum=0;

    msgnum = summary_get_current_msgnum(mainwin->summaryview);
    folderview_unselect(mainwin->folderview);
    folderview_select(mainwin->folderview,mainwin->summaryview->cur_folder);
    if (prefs_common.no_moving_inc) summary_select_by_msgnum(mainwin->summaryview, msgnum);

};
void inc_mail(MainWindow *mainwin)
{
    gint msgnum=0;
    if (prefs_common.use_extinc && prefs_common.extinc_path) {
		gint pid;

		/* external incorporating program */
		if ((pid = fork()) < 0) {
			perror("fork");
			return;
		}

		if (pid == 0) {
			execlp(prefs_common.extinc_path,
			       g_basename(prefs_common.extinc_path),
			       NULL);

			/* this will be called when failed */
			perror("exec");
			_exit(1);
		}

		/* wait until child process is terminated */
		waitpid(pid, NULL, 0);

		if (prefs_common.inc_local) inc_spool();
	} else {
		if (prefs_common.inc_local) inc_spool();

		inc_account_mail(cur_account, mainwin);
	}
    inc_update(mainwin);
}


void inc_all_account_mail(MainWindow *mainwin)
{
	if (prefs_common.inc_local) inc_spool();

	account_foreach(inc_account_mail, mainwin);
        inc_update(mainwin);
}

gint inc_all_account_mail_by_timer(MainWindow *mainwin)
{
#if USE_THREADS
	gdk_threads_enter();
#endif
	if (prefs_common.inc_local) inc_spool();
	account_foreach(inc_account_mail_by_timer, mainwin);
        inc_update(mainwin);
#if USE_THREADS
	gdk_threads_leave();
#endif
	return TRUE;
}

static void pop3_automaton_create(PrefsAccount *ac_prefs, MainWindow *mainwin,
				  IncState *retval)
{
	Automaton *atm;
	Pop3State *state;
	SockInfo *sockinfo;
	gint i;
	gchar *server;
	gushort port;
	gchar *buf;

	atm = automaton_create(N_POP3_PHASE);

	atm->data = state = g_new0(Pop3State, 1);
	state->ac_prefs = ac_prefs;
	state->folder_table = g_hash_table_new(g_str_hash, g_str_equal);
	state->mainwin = mainwin;
	state->id_table = inc_get_uidl_table(ac_prefs);
	state->id_list = NULL;
	state->new_id_list = NULL;
	state->retval = retval;

	state->user = g_strdup(ac_prefs->userid);

	if (ac_prefs->passwd)
		state->pass = g_strdup(ac_prefs->passwd);
	else if (ac_prefs->tmp_pass)
		state->pass = g_strdup(ac_prefs->tmp_pass);
	else {
		gchar *pass;
		gchar *message;

		message = g_strdup_printf(_("Input password for %s:"),
					  state->user);

		pass = input_dialog_with_invisible(_("Input password"),
						   message, NULL);
		manage_window_focus_in(mainwin->window, NULL, NULL);
		if (pass) {
			ac_prefs->tmp_pass = g_strdup(pass);
			state->pass = pass;
		} else {
			pop3_terminate(-1, atm);
			return;
		}

		g_free(message);
	}

	state->progress = inc_progress_create(atm);
	buf = g_strdup_printf(_("%s: Retrieving new messages"),
			      ac_prefs->recv_server);
	gtk_window_set_title(GTK_WINDOW(state->progress->window), buf);
	g_free(buf);

	atm->state[POP3_GREETING_RECV].handler = pop3_greeting_recv;
	atm->state[POP3_GETAUTH_USER_SEND].handler = pop3_getauth_user_send;
	atm->state[POP3_GETAUTH_USER_RECV].handler = pop3_getauth_user_recv;
	atm->state[POP3_GETAUTH_PASS_SEND].handler = pop3_getauth_pass_send;
	atm->state[POP3_GETAUTH_PASS_RECV].handler = pop3_getauth_pass_recv;
	atm->state[POP3_GETAUTH_APOP_SEND].handler = pop3_getauth_apop_send;
	atm->state[POP3_GETAUTH_APOP_RECV].handler = pop3_getauth_apop_recv;
	atm->state[POP3_GETRANGE_STAT_SEND].handler = pop3_getrange_stat_send;
	atm->state[POP3_GETRANGE_STAT_RECV].handler = pop3_getrange_stat_recv;
	atm->state[POP3_GETRANGE_LAST_SEND].handler = pop3_getrange_last_send;
	atm->state[POP3_GETRANGE_LAST_RECV].handler = pop3_getrange_last_recv;
	atm->state[POP3_GETRANGE_UIDL_SEND].handler = pop3_getrange_uidl_send;
	atm->state[POP3_GETRANGE_UIDL_RECV].handler = pop3_getrange_uidl_recv;
	atm->state[POP3_RETR_SEND].handler = pop3_retr_send;
	atm->state[POP3_RETR_RECV].handler = pop3_retr_recv;
	atm->state[POP3_DELETE_SEND].handler = pop3_delete_send;
	atm->state[POP3_DELETE_RECV].handler = pop3_delete_recv;
	atm->state[POP3_LOGOUT_SEND].handler = pop3_logout_send;
	atm->state[POP3_LOGOUT_RECV].handler = pop3_logout_recv;
	atm->state[POP3_GREETING_RECV].condition = GDK_INPUT_READ;
	for (i = POP3_GETAUTH_USER_SEND; i < N_POP3_PHASE; ) {
		atm->state[i++].condition = GDK_INPUT_WRITE;
		atm->state[i++].condition = GDK_INPUT_READ;
	}
	atm->terminate = (AtmHandler)pop3_terminate;

	atm->num = POP3_GREETING_RECV;

	server = ac_prefs->recv_server;
	port = ac_prefs->set_popport ? ac_prefs->popport : 110;

	buf = g_strdup_printf(_("Connecting to POP3 server: %s ..."), server);
	log_message("%s\n", buf);
	progress_dialog_set_label(state->progress, buf);
	g_free(buf);
	GTK_EVENTS_FLUSH();
#if USE_THREADS
	sockinfo = g_new0(SockInfo, 1);
	sockinfo->hostname = g_strdup(server);
	sockinfo->port = port;
	sockinfo->sock = -2;
	sockinfo->state = CONN_READY;
	sockinfo->timer = time(NULL);
	sockinfo->data = atm;
	state->sockinfo = sockinfo;
	atm->timeout_tag = gtk_timeout_add(TIMEOUT_ITV,
	                                   (GtkFunction)connection_check_cb, atm);
	pthread_create(&sockinfo->connect_thr, NULL,
	               (void *)pop3_receive_threaded, sockinfo);
#else
	if ((sockinfo = sock_connect_nb(server, port)) == NULL) {
		log_warning(_("Can't connect to POP3 server: %s:%d\n"),
			    server, port);
		manage_window_focus_in(state->progress->window, NULL, NULL);
		alertpanel_error(_("Can't connect to POP3 server: %s:%d"),
				 server, port);
		manage_window_focus_out(state->progress->window, NULL, NULL);
		pop3_terminate(-1, atm);
		return;
	}
	state->sockinfo = sockinfo;
	sockinfo->data = atm;

	atm->tag = gdk_input_add(sockinfo->sock,
				 atm->state[atm->num].condition,
				 automaton_input_cb, atm);
#endif
	gtk_main();
	sock_sockinfo_free(sockinfo);
}

static void pop3_automaton_destroy(Automaton *atm)
{
	Pop3State *state = atm->data;

	g_hash_table_destroy(state->folder_table);
	if (state->id_table) {
		hash_free_strings(state->id_table);
		g_hash_table_destroy(state->id_table);
	}
	slist_free_strings(state->id_list);
	slist_free_strings(state->new_id_list);
	g_slist_free(state->id_list);
	g_slist_free(state->new_id_list);

	g_free(state->greeting);
	g_free(state->user);
	g_free(state->pass);
	g_free(state->prev_folder);

	g_free(state);

	automaton_destroy(atm);
}

static gint pop3_terminate(gint source, Automaton *atm)
{
	Pop3State *state = atm->data;

	if (atm->tag > 0) gdk_input_remove(atm->tag);
	if (atm->timeout_tag > 0) gtk_timeout_remove(atm->timeout_tag);
	if (source > 0) sock_close(source);

	if (state->error_val == PS_AUTHFAIL) {
		manage_window_focus_in(state->progress->window, NULL, NULL);
		alertpanel_error(_("Authorization for %s failed"),
				 state->user);
	}

	statusbar_pop_all();
	progress_dialog_destroy(state->progress);
	manage_window_focus_in(state->mainwin->window, NULL, NULL);

	folderview_scan_folder_foreach_a(state->folder_table);	

	if (state->error_val == PS_AUTHFAIL && state->ac_prefs->tmp_pass) {
		g_free(state->ac_prefs->tmp_pass);
		state->ac_prefs->tmp_pass = NULL;
	}

	inc_write_uidl_list(state);

	pop3_automaton_destroy(atm);

	if (source > 0 || source == -2) gtk_main_quit();

	return 0;
}

static GHashTable *inc_get_uidl_table(PrefsAccount *ac_prefs)
{
	GHashTable *table;
	gchar *path;
	FILE *fp;
	gchar buf[IDLEN + 3];

	path = g_strconcat(get_rc_dir(), G_DIR_SEPARATOR_S,
			   "uidl-", ac_prefs->recv_server,
			   "-", ac_prefs->userid, NULL);
	if ((fp = fopen(path, "r")) == NULL) {
		if (ENOENT != errno) FILE_OP_ERROR(path, "fopen");
		g_free(path);
		return NULL;
	}
	g_free(path);

	table = g_hash_table_new(g_str_hash, g_str_equal);

	while (fgets(buf, sizeof(buf), fp) != NULL) {
		strretchomp(buf);
		g_hash_table_insert(table, g_strdup(buf), GINT_TO_POINTER(1));
	}

	fclose(fp);

	return table;
}

static void inc_write_uidl_list(Pop3State *state)
{
	gchar *path;
	FILE *fp;
	GSList *cur;

	if (!state->id_list) return;

	path = g_strconcat(get_rc_dir(), G_DIR_SEPARATOR_S,
			   "uidl-", state->ac_prefs->recv_server,
			   "-", state->user, NULL);
	if ((fp = fopen(path, "w")) == NULL) {
		FILE_OP_ERROR(path, "fopen");
		g_free(path);
		return;
	}

	for (cur = state->id_list; cur != NULL; cur = cur->next) {
		if (fputs((gchar *)cur->data, fp) == EOF) {
			FILE_OP_ERROR(path, "fputs");
			break;
		}
		if (fputc('\n', fp) == EOF) {
			FILE_OP_ERROR(path, "fputc");
			break;
		}
	}

	if (fclose(fp) == EOF) FILE_OP_ERROR(path, "fclose");
	g_free(path);
}

#if USE_THREADS

static void pop3_receive_threaded(SockInfo *sockinfo)
{
	Automaton *atm = (Automaton *)sockinfo->data;
	SockInfo *new_sockinfo;
	gint next;

	new_sockinfo = sock_connect(sockinfo->hostname, sockinfo->port);
	if (!new_sockinfo) {
		sockinfo->state = CONN_FAILED;
		pthread_exit(0);
	}
	g_free(new_sockinfo->hostname);
	sockinfo->sock = new_sockinfo->sock;
	sockinfo->state = new_sockinfo->state;
	sockinfo->timer = time(NULL);
	
	g_free(new_sockinfo);
	while (1) {
		next = atm->state[atm->num].handler(sockinfo->sock, atm->data);
		if (next >= 0 && next <= atm->max && next != atm->num) {
			atm->num = next;
			continue;
		}
		break;
	}
	sockinfo->state = CONN_DONE;
}

static gint connection_check_cb(Automaton *atm)
{
	Pop3State *state = atm->data;
	SockInfo *sockinfo = state->sockinfo;

	gdk_threads_enter();

	if (sockinfo->state == CONN_LOOKUPFAILED ||
	    sockinfo->state == CONN_FAILED) {
		log_warning(_("Can't connect to POP3 server: %s:%d\n"),
			    sockinfo->hostname, sockinfo->port);
		manage_window_focus_in(state->progress->window, NULL, NULL);
		
		alertpanel_error(_("Can't connect to POP3 server: %s:%d"),
				 sockinfo->hostname, sockinfo->port);
		manage_window_focus_out(state->progress->window, NULL, NULL);
		pop3_terminate(sockinfo->sock, atm);
		gdk_threads_leave();
		return FALSE;
	} else if (sockinfo->state == CONN_READY ||
	            sockinfo->state == CONN_LOOKUPSUCCESS) {
		if (time(NULL) > sockinfo->timer+CONNECT_TIMEOUT) {
			log_warning(_("Connection timeout\n"));
			manage_window_focus_in(state->progress->window, NULL, NULL);
			
			alertpanel_error(_("Connection timeout\n"));
			
			pthread_cancel(sockinfo->connect_thr);
			pop3_terminate(sockinfo->sock, atm);
			gdk_threads_leave();
			return FALSE;
		}
		gdk_threads_leave();
		return TRUE;
	} else if (sockinfo->state == CONN_ESTABLISHED) {
		if (time(NULL) > sockinfo->timer+RECEIVE_TIMEOUT) {
			log_warning(_("Socket read timeout\n"));
			manage_window_focus_in(state->progress->window, NULL, NULL);
			
			alertpanel_error(_("Socket read timeout\n"));
			
			pthread_cancel(sockinfo->connect_thr);
			pop3_terminate(sockinfo->sock, atm);
			gdk_threads_leave();
			return FALSE;
		}
		gdk_threads_leave();
		return TRUE;
	} else if (sockinfo->state == CONN_DONE) {
		atm->terminate(sockinfo->sock, atm);
		gdk_threads_leave();
		return FALSE;
	} else {
		gdk_threads_leave();
		return TRUE;
	}
}
#endif

void inc_progress_update(Pop3State *state, Pop3Phase phase)
{
	gchar buf[MSGBUFSIZE];

	switch (phase) {
	case POP3_GREETING_RECV:
		break;
	case POP3_GETAUTH_USER_SEND:
	case POP3_GETAUTH_USER_RECV:
	case POP3_GETAUTH_PASS_SEND:
	case POP3_GETAUTH_PASS_RECV:
	case POP3_GETAUTH_APOP_SEND:
	case POP3_GETAUTH_APOP_RECV:
		progress_dialog_set_label(state->progress, _("Authorizing"));
		break;
	case POP3_GETRANGE_STAT_SEND:
	case POP3_GETRANGE_STAT_RECV:
	case POP3_GETRANGE_LAST_SEND:
	case POP3_GETRANGE_LAST_RECV:
	case POP3_GETRANGE_UIDL_SEND:
	case POP3_GETRANGE_UIDL_RECV:
		progress_dialog_set_label(state->progress,
					  _("Getting number of new messages"));
		break;
	case POP3_RETR_SEND:
	case POP3_RETR_RECV:
		g_snprintf(buf, sizeof(buf),
			   _("Retrieving message (%d / %d)"),
			   state->cur_msg, state->count);
		progress_dialog_set_label(state->progress, buf);
		progress_dialog_set_percentage
			(state->progress,
			 (gfloat)state->cur_msg / state->count);
		break;
	case POP3_DELETE_SEND:
	case POP3_DELETE_RECV:
		progress_dialog_set_label(state->progress,
					  _("Deleting message"));
		break;
	case POP3_LOGOUT_SEND:
	case POP3_LOGOUT_RECV:
		progress_dialog_set_label(state->progress, _("Quitting"));
		break;
	default:
	}
}

gint inc_drop_message(const gchar *file, Pop3State *state)
{
	gchar *inbox;
	gchar *dropfolder;
	gchar *path;
	gint msgnum;
	gint val;

	if (state->ac_prefs->inbox && is_dir_exist(state->ac_prefs->inbox))
		inbox = state->ac_prefs->inbox;
	else
		inbox = INBOX_DIR;

	if (state->ac_prefs->filter_on_recv) {
		dropfolder =
			filter_get_dest_folder(prefs_common.fltlist, file);
		if (!dropfolder)
			dropfolder = inbox;
		else if (!strcmp(dropfolder, FILTER_NOT_RECEIVE)) {
			g_warning(_("a message isn't received\n"));
			return 1;
		} else if (!is_dir_exist(dropfolder))
			dropfolder = inbox;
	} else
		dropfolder = inbox;

	msgnum = GPOINTER_TO_INT(g_hash_table_lookup
				 (state->folder_table, dropfolder));
	if (!msgnum) {
		msgnum = procmsg_get_last_message_number(dropfolder);
		if (msgnum < 0) {
			g_warning(_("can't drop message into %s\n"),
				  dropfolder);
			return -1;
		}
		g_hash_table_insert(state->folder_table, g_strdup(dropfolder),
				    GINT_TO_POINTER(msgnum + 1));
	} else
		g_hash_table_insert(state->folder_table, dropfolder,
				    GINT_TO_POINTER(msgnum + 1));
	msgnum++;

	path = g_strdup_printf("%s%c%d", dropfolder, G_DIR_SEPARATOR, msgnum);

	if (is_file_exist(path)) {
		g_warning(_("%s exists\n"), path);
		val = -1;
	} else if ((val = rename(file, path)) < 0) {
		FILE_OP_ERROR(file, "rename");
		g_warning("can't move tmpmsg to %s\n", path);
	}

	g_free(path);
	if (val < 0) unlink(file);

	return val;
}

static gint inc_account_mail(PrefsAccount *ac_prefs, gpointer data)
{
	MainWindow *mainwin = (MainWindow *)data;
	IncState retval = INC_SUCCESS;

	if (!ac_prefs) return -1;
	if ((ac_prefs->protocol == A_POP3 ||
	     ac_prefs->protocol == A_APOP) &&
	     ac_prefs->recv_server) {
		debug_print(_("getting new messages of account %s...\n"),
			    ac_prefs->account_name);

		pop3_automaton_create(ac_prefs, mainwin, &retval);
	}

	switch (retval) {
	case INC_ERROR:
		alertpanel_error(_("Error occured while processing mail."));
		break;
	case INC_NOSPACE:
		alertpanel_error(_("No disk space left."));
		break;
	default:
	}

	return retval;
}

static gint inc_account_mail_by_timer(PrefsAccount *ac_prefs, gpointer data)
{
	MainWindow *mainwin = (MainWindow *)data;
	IncState retval = INC_SUCCESS;
	if (!ac_prefs) return -1;
	if (!ac_prefs->autochk) return retval;
	if ((ac_prefs->protocol == A_POP3 ||
	     ac_prefs->protocol == A_APOP) &&
	     ac_prefs->recv_server) {
		debug_print(_("getting new messages of account %s...\n"),
			    ac_prefs->account_name);

		pop3_automaton_create(ac_prefs, mainwin, &retval);
	}

	switch (retval) {
	case INC_ERROR:
		alertpanel_error(_("Error occured while processing mail."));
		break;
	case INC_NOSPACE:
		alertpanel_error(_("No disk space left."));
		break;
	default:
	}

	return retval;
}

static ProgressDialog *inc_progress_create(gpointer data)
{
	ProgressDialog *progress;

	progress = progress_dialog_create();
	gtk_window_set_title(GTK_WINDOW(progress->window),
			     _("Retrieving new messages"));
	gtk_signal_connect(GTK_OBJECT(progress->cancel_btn), "clicked",
			   GTK_SIGNAL_FUNC(inc_cancel), data);
	gtk_signal_connect(GTK_OBJECT(progress->window), "delete_event",
			   GTK_SIGNAL_FUNC(gtk_true), NULL);
	manage_window_set_transient(GTK_WINDOW(progress->window));

	progress_dialog_set_value(progress, 0.0);
	if (prefs_common.show_progress)	gtk_widget_show_now(progress->window);

	return progress;
}

static void inc_cancel(GtkWidget *widget, gpointer data)
{
	Automaton *atm = (Automaton *)data;
	Pop3State *state = atm->data;
	SockInfo *sockinfo = state->sockinfo;

#if USE_THREADS
	if (sockinfo->state == CONN_READY ||
	    sockinfo->state == CONN_LOOKUPSUCCESS ||
	    sockinfo->state == CONN_ESTABLISHED) {
		pthread_cancel(sockinfo->connect_thr);
		//pthread_kill(sockinfo->connect_thr, SIGINT);
		g_print("connection was cancelled.\n");
	}
#endif

	*state->retval = INC_CANCEL;
	pop3_terminate(sockinfo->sock, atm);
}

static gint inc_spool(void)
{
	gchar *mbox, *logname;
	gint msgs;

	logname = g_get_user_name();
	mbox = g_strconcat(prefs_common.spool_path
			   ? prefs_common.spool_path : DEFAULT_SPOOL_PATH,
			   G_DIR_SEPARATOR_S, logname, NULL);
	msgs = get_spool(NULL, mbox);
	g_free(mbox);

	return msgs;
}

static gint get_spool(const gchar *dest, const gchar *mbox)
{
	gint msgs, size;
	gint lockfd;
	gchar *tmp_mbox = "/tmp/tmpmbox";
	GHashTable *folder_table = NULL;

	g_return_val_if_fail(mbox != NULL, -1);
	CHDIR_RETURN_VAL_IF_FAIL(maildir, -1);

	if (!dest) dest = INBOX_DIR;

	if (!is_file_exist(mbox) || (size = get_file_size(mbox)) == 0) {
		debug_print(_("no messages in local mailbox.\n"));
		return 0;
	} else if (size < 0)
		return -1;

	if ((lockfd = lock_mbox(mbox, LOCK_FLOCK)) < 0)
		return -1;

	if (copy_mbox(mbox, tmp_mbox) < 0)
		return -1;

	debug_print(_("Getting new messages from %s into %s...\n"),
		    mbox, dest);

	if (prefs_common.filter_on_inc)
		folder_table = g_hash_table_new(g_str_hash, g_str_equal);
	msgs = proc_mbox(dest, tmp_mbox, folder_table);

	unlink(tmp_mbox);
	if (msgs >= 0) empty_mbox(mbox);
	unlock_mbox(mbox, lockfd, LOCK_FLOCK);

	if (folder_table) {
		folderview_scan_folder_foreach_a(folder_table);
		g_hash_table_destroy(folder_table);
	} else
		folderview_scan_folder_a(dest, FALSE);
	return msgs;
}
