/*
    Stuphead: (C) 2000,2001 Grigroy Bakunov, Sergey Pinaev
 */
/*
 * Sylpheed -- a GTK+ based, lightweight, and fast e-mail client
 * Copyright (C) 1999,2000 Hiroyuki Yamamoto
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <glib.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtkvpaned.h>
#include <gtk/gtkcontainer.h>
#include <gtk/gtkeditable.h>
#include <gtk/gtkwindow.h>
#include <gtk/gtktext.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include "main.h"
#include "messageview.h"
#include "headerview.h"
#include "textview.h"
#include "imageview.h"
#include "mimeview.h"
#include "htmlview.h"
#include "procmsg.h"
#include "procmime.h"
#include "procheader.h"
#include "prefs_common.h"
#include "utils.h"

static void messageview_change_view_type(MessageView	*messageview,
					 MessageType	 type);
static void messageview_destroy_cb	(GtkWidget	*widget,
					 MessageView	*messageview);
static void messageview_size_allocate_cb(GtkWidget	*widget,
					 GtkAllocation	*allocation);
static void key_pressed			(GtkWidget	*widget,
					 GdkEventKey	*event,
					 MessageView	*messageview);

MessageView *messageview_create(void)
{
	MessageView *messageview;
	GtkWidget *paned;
	HeaderView *headerview;
	TextView *textview;
	ImageView *imageview;
	MimeView *mimeview;
	HtmlView *htmlview;


	debug_print(_("Creating message view...\n"));
	messageview = g_new0(MessageView, 1);

	messageview->type = MVIEW_TEXT;

	headerview = headerview_create();

	textview = textview_create();
	textview->messageview = messageview;

	imageview = imageview_create();
	imageview->messageview = messageview;

	htmlview = htmlview_create();
	htmlview->messageview = messageview;

	mimeview = mimeview_create();
	mimeview->textview = textview;
	mimeview->imageview = imageview;
	mimeview->htmlview = htmlview;
	mimeview->messageview = messageview;

	/* to remove without destroyed */
	gtk_widget_ref(*(GtkWidget **)textview);
	gtk_widget_ref(*(GtkWidget **)imageview);
	gtk_widget_ref(*(GtkWidget **)htmlview);
	gtk_widget_ref(*(GtkWidget **)mimeview);

	paned = gtk_vpaned_new();
	messageview->paned      = paned;
	messageview->new_window = FALSE;
	messageview->window     = NULL;
	messageview->headerview = headerview;
	messageview->textview   = textview;
	messageview->imageview  = imageview;
	messageview->htmlview  	= htmlview;
	messageview->mimeview   = mimeview;
	return messageview;
}

MessageView *messageview_create_with_new_window(void)
{
	GtkWidget *window;
	MessageView *msgview;

	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(window), "Message");
	gtk_window_set_policy(GTK_WINDOW(window), TRUE, TRUE, FALSE);
	gtk_widget_set_usize(window, prefs_common.msgwin_width,
			     prefs_common.msgwin_height);

	msgview = messageview_create();

	gtk_signal_connect(GTK_OBJECT(window), "size_allocate",
			   GTK_SIGNAL_FUNC(messageview_size_allocate_cb),
			   msgview);
	gtk_signal_connect(GTK_OBJECT(window), "destroy",
			   GTK_SIGNAL_FUNC(messageview_destroy_cb), msgview);
	gtk_signal_connect(GTK_OBJECT(window), "key_press_event",
			   GTK_SIGNAL_FUNC(key_pressed), msgview);

	gtk_container_add(GTK_CONTAINER(window), *(GtkWidget **)msgview);
	gtk_widget_show_all(window);

	msgview->new_window = TRUE;
	msgview->window = window;

	messageview_init(msgview);

	return msgview;
}

void messageview_init(MessageView *messageview)
{
	TextView *textview = messageview->textview;
	ImageView *imageview = messageview->imageview;
	HtmlView *htmlview = messageview->htmlview;
	MimeView *mimeview = messageview->mimeview;
	gtk_paned_add1(GTK_PANED(messageview->paned), *(GtkWidget **)messageview->headerview);
#ifdef USE_MOZILLA
	gtk_paned_add2(GTK_PANED(messageview->paned), *(GtkWidget **)messageview->mimeview);
#else
	gtk_paned_add2(GTK_PANED(messageview->paned), *(GtkWidget **)textview);
#endif

	htmlview_init(messageview->htmlview);
	mimeview_init(messageview->mimeview);
	textview_init(messageview->textview);
	imageview_init(messageview->imageview);
	headerview_init(messageview->headerview);
	messageview_set_font();
}

void messageview_show(MessageView *messageview, MsgInfo *msginfo)
{
	FILE *fp;
	MimeInfo *mimeinfo;
	MsgInfo *tmpmsginfo;
	gchar *file;
	
	g_return_if_fail(msginfo != NULL);

	if ((fp = procmsg_open_message(msginfo)) == NULL) return;
	mimeinfo = procmime_scan_mime_header(fp);
	fclose(fp);
	if (!mimeinfo) return;

	file = (gchar*) procmsg_get_message_file_path(msginfo);
	g_return_if_fail(file != NULL);

	tmpmsginfo = procheader_parse(file, msginfo->flags, TRUE);
	headerview_show(messageview->headerview, tmpmsginfo);
	procmsg_msginfo_free(tmpmsginfo);

#ifdef USE_MOZILLA
	if (mimeinfo->mime_type != MIME_TEXT) {
		gtk_widget_set_usize(messageview->mimeview->scrolledwin, -1, 80);
		messageview_change_view_type(messageview, MVIEW_MIME);
	} else {
		gtk_widget_set_usize(messageview->mimeview->scrolledwin, -1, 1);
		messageview_change_view_type(messageview, MVIEW_TEXT);
	}
	mimeview_show_message(messageview->mimeview, mimeinfo, file);
#else
	if (mimeinfo->mime_type != MIME_TEXT) {
		messageview_change_view_type(messageview, MVIEW_MIME);
		mimeview_show_message(messageview->mimeview, mimeinfo, file);
	} else {
		messageview_change_view_type(messageview, MVIEW_TEXT);
		textview_show_message(messageview->textview, mimeinfo, file);
		procmime_mimeinfo_free(mimeinfo);
	}
#endif
	g_free(file);
	
}

static void messageview_change_view_type(MessageView *messageview,
					 MessageType type)
{

	TextView *textview = messageview->textview;
	ImageView *imageview = messageview->imageview;
	HtmlView *htmlview = messageview->htmlview;
	MimeView *mimeview = messageview->mimeview;


	if (messageview->type == type) return;

#ifdef USE_MOZILLA
	messageview->type = type;
#else
	if (type == MVIEW_MIME) {
                gtk_container_remove(GTK_CONTAINER(*(GtkWidget **)messageview),
                                     *(GtkWidget **)textview);
		gtk_paned_add2(GTK_PANED(messageview->paned),
                               *(GtkWidget **)mimeview);
                gtk_container_add(GTK_CONTAINER(mimeview->vbox),
				  *(GtkWidget **)textview);
		mimeview->type = MIMEVIEW_TEXT;
	} else if (type == MVIEW_TEXT) {

                gtk_container_remove(GTK_CONTAINER(*(GtkWidget **)messageview),
				     *(GtkWidget **)mimeview);

		if (mimeview->vbox == (*(GtkWidget **)textview)->parent) {
                        gtk_container_remove(GTK_CONTAINER(mimeview->vbox),
					     *(GtkWidget **)textview);
		} else if (mimeview->vbox == (*(GtkWidget**)htmlview)->parent){
                        gtk_container_remove(GTK_CONTAINER(mimeview->vbox),
					      *(GtkWidget**)htmlview);
		} else if (mimeview->vbox == (*(GtkWidget**)imageview)->parent){
			gtk_container_remove(GTK_CONTAINER(mimeview->vbox),
					      *(GtkWidget**)imageview);
		} 

		gtk_paned_add2(GTK_PANED(messageview->paned),
                               *(GtkWidget **)textview);
	} else
		return;
	messageview->type = type;
#endif
}

void messageview_clear(MessageView *messageview)
{
	debug_print("Clearing message view...");
	mimeview_change_view_type(messageview->mimeview, MVIEW_TEXT);
	messageview_change_view_type(messageview, MVIEW_TEXT);
	mimeview_clear(messageview->mimeview);
	headerview_clear(messageview->headerview);
	textview_clear(messageview->textview);
	debug_print(" done\n");
}

void messageview_destroy(MessageView *messageview)
{
	GtkWidget *textview  = *(GtkWidget **)messageview->textview;
	GtkWidget *imageview = *(GtkWidget **)messageview->imageview;
	GtkWidget *mimeview  = *(GtkWidget **)messageview->mimeview;
	GtkWidget *htmlview = *(GtkWidget **)messageview->htmlview;
	textview_destroy(messageview->textview);
	imageview_destroy(messageview->imageview);
	htmlview_destroy(messageview->htmlview);
	mimeview_destroy(messageview->mimeview);

	g_free(messageview);

	gtk_widget_unref(textview);
	gtk_widget_unref(imageview);
	gtk_widget_unref(htmlview);
	gtk_widget_unref(mimeview);
}

void messageview_quote_color_set(void)
{
}

void messageview_set_font(void)
{
	textview_set_font();
}

void messageview_copy_clipboard(MessageView *messageview)
{
	if (messageview->type == MVIEW_TEXT)
		gtk_editable_copy_clipboard
			(GTK_EDITABLE(messageview->textview->text));
}

void messageview_select_all(MessageView *messageview)
{
	if (messageview->type == MVIEW_TEXT)
		gtk_editable_select_region
			(GTK_EDITABLE(messageview->textview->text), 0, -1);
}

GtkWidget *messageview_get_text_widget(MessageView *messageview)
{
	return messageview->textview->text;
}

static void messageview_destroy_cb(GtkWidget *widget, MessageView *messageview)
{
	messageview_destroy(messageview);
}

static void messageview_size_allocate_cb(GtkWidget *widget,
					 GtkAllocation *allocation)
{
	g_return_if_fail(allocation != NULL);

	prefs_common.msgwin_width  = allocation->width;
	prefs_common.msgwin_height = allocation->height;
}

static void key_pressed(GtkWidget *widget, GdkEventKey *event,
			MessageView *messageview)
{
	if (event && event->keyval == GDK_Escape && messageview->window)
		gtk_widget_destroy(messageview->window);
}
