/*
    Stuphead: (C) 2000,2001 Grigroy Bakunov, Sergey Pinaev
 */
/*
 * Sylpheed -- a GTK+ based, lightweight, and fast e-mail client
 * Copyright (C) 1999,2000 Hiroyuki Yamamoto
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <glib.h>
#include <stdio.h>
#include <stdlib.h>
#include <dirent.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>

#undef MEASURE_TIME

#ifdef MEASURE_TIME
#  include <sys/time.h>
#endif

#include "main.h"
#include "utils.h"
#include "procmsg.h"
#include "procheader.h"
#include "send.h"
#include "procmime.h"

typedef struct _FlagInfo	FlagInfo;

struct _FlagInfo
{
	guint    msgnum;
	MsgFlags flags;
};

static GSList *procmsg_get_uncached_msgs	(GHashTable	*msg_table,
						 const gchar	*folder);

static void mark_sum_func			(gpointer	 key,
						 gpointer	 value,
						 gpointer	 data);

static GHashTable *procmsg_read_mark_file	(const gchar	*folder);
static gint procmsg_cmp_msgnum			(gconstpointer	 a,
						 gconstpointer	 b);
static gint procmsg_cmp_msgnum_for_sort		(gconstpointer	 a,
						 gconstpointer	 b);
static gint procmsg_cmp_flag_msgnum		(gconstpointer	 a,
						 gconstpointer	 b);

GSList *procmsg_get_msgs(const gchar *folder, gboolean use_cache,
			 gboolean scan_new)
{
	GSList *mlist;
	GHashTable *msg_table;
#ifdef MEASURE_TIME
	struct timeval tv_before, tv_after, tv_result;

	gettimeofday(&tv_before, NULL);
#endif

	if (use_cache && !scan_new) {
		mlist = procmsg_read_cache(folder, M_MAIL, FALSE);
		if (!mlist)
			mlist = procmsg_get_uncached_msgs(NULL, folder);
	} else if (use_cache) {
		GSList *newlist;

		mlist = procmsg_read_cache(folder, M_MAIL, TRUE);
		msg_table = procmsg_msg_hash_table_create(mlist);

		newlist = procmsg_get_uncached_msgs(msg_table, folder);
		if (msg_table)
			g_hash_table_destroy(msg_table);

		mlist = g_slist_concat(mlist, newlist);
	} else
		mlist = procmsg_get_uncached_msgs(NULL, folder);

	procmsg_set_flags(mlist, folder, M_MAIL);

#ifdef MEASURE_TIME
	gettimeofday(&tv_after, NULL);

	timersub(&tv_after, &tv_before, &tv_result);
	g_print("procmsg_get_msgs: %s: elapsed time: %ld.%06ld sec\n",
		folder, tv_result.tv_sec, tv_result.tv_usec);
#endif

	return mlist;
}

GHashTable *procmsg_msg_hash_table_create(GSList *mlist)
{
	GHashTable *msg_table;

	if (mlist == NULL) return NULL;

	msg_table = g_hash_table_new(NULL, g_direct_equal);
	procmsg_msg_hash_table_append(msg_table, mlist);

	return msg_table;
}

void procmsg_msg_hash_table_append(GHashTable *msg_table, GSList *mlist)
{
	GSList *cur;
	MsgInfo *msginfo;

	if (msg_table == NULL || mlist == NULL) return;

	for (cur = mlist; cur != NULL; cur = cur->next) {
		msginfo = (MsgInfo *)cur->data;

		g_hash_table_insert(msg_table,
				    GUINT_TO_POINTER(msginfo->msgnum),
				    msginfo);
	}
}

GSList *procmsg_read_cache(const gchar *folder, MsgType type,
			   gboolean scan_file)
{
	GSList *mlist = NULL;
	GSList *last = NULL;
	FILE *fp;
	gchar buf[BUFFSIZE];
        gchar file_buf[BUFFSIZE];
        MsgInfo *msginfo;
	MsgFlags default_flags;	
	struct stat s;
	gint ver;
	guint num;
	guint len;
	g_return_val_if_fail(folder != NULL, NULL);

	default_flags = MSG_NEW|MSG_UNREAD|MSG_CACHED;
	if (type == M_MAIL) {
		if (!strcmp(folder, QUEUE_DIR)) {
			MSG_SET_FLAGS(default_flags, MSG_QUEUED);
		} else if (!strcmp(folder, DRAFT_DIR)) {
			MSG_SET_FLAGS(default_flags, MSG_DRAFT);
	    }
	} else if (type == M_IMAP) {
		MSG_SET_FLAGS(default_flags, MSG_IMAP);
	} else if (type == M_NEWS) {
		MSG_SET_FLAGS(default_flags, MSG_NEWS);
	}

	if (type == M_MAIL) {
		CHDIR_RETURN_VAL_IF_FAIL(maildir, NULL);
		CHDIR_RETURN_VAL_IF_FAIL(folder, NULL);

		if ((fp = fopen(CACHE_FILE, "r")) == NULL) {
			debug_print(_("\tNo cache file\n"));
			CHDIR_RETURN_VAL_IF_FAIL(maildir, NULL);
			return NULL;
		}
	} else {
		gchar *cache_file;

		cache_file = g_strconcat(get_news_cache_dir(),
					 G_DIR_SEPARATOR_S, folder,
					 G_DIR_SEPARATOR_S, CACHE_FILE, NULL);
		if ((fp = fopen(cache_file, "r")) == NULL) {
			debug_print(_("\tNo cache file\n"));
			g_free(cache_file);
			return NULL;
		}
                setvbuf(fp, file_buf, _IOFBF, sizeof (file_buf));
                g_free(cache_file);
	}

	debug_print(_("\tReading summary cache..."));

	/* compare cache version */
	if (fread(&ver, sizeof(ver), 1, fp) != 1 ||
	    CACHE_VERSION != ver) {
		debug_print(_("Cache version is different. Discarding it.\n"));
		fclose(fp);
		CHDIR_RETURN_VAL_IF_FAIL(maildir, NULL);
		return NULL;
	}

	while (fread(&num, sizeof(num), 1, fp) == 1) {
		msginfo = g_new0(MsgInfo, 1);
		msginfo->msgnum = num;
		READ_CACHE_DATA_INT(msginfo->size, fp);
		READ_CACHE_DATA_INT(msginfo->mtime, fp);
		READ_CACHE_DATA_INT(msginfo->date_t, fp);
		READ_CACHE_DATA_INT(msginfo->flags, fp);

		READ_CACHE_DATA(msginfo->fromname, fp);

		READ_CACHE_DATA(msginfo->date, fp);
		READ_CACHE_DATA(msginfo->from, fp);
		READ_CACHE_DATA(msginfo->to, fp);
		READ_CACHE_DATA(msginfo->newsgroups, fp);
		READ_CACHE_DATA(msginfo->subject, fp);
		READ_CACHE_DATA(msginfo->msgid, fp);
		READ_CACHE_DATA(msginfo->inreplyto, fp);
		READ_CACHE_DATA(msginfo->default_mime_charset, fp);

		MSG_SET_FLAGS(msginfo->flags, default_flags);

		if (type == M_MAIL)
			msginfo->flags |= MSG_NEW|MSG_UNREAD|MSG_CACHED;
		else if (type == M_NEWS)
			msginfo->flags |=
				MSG_NEW|MSG_UNREAD|MSG_NEWS|MSG_CACHED;

		/* if the message file doesn't exist or is changed,
		   don't add the data */
		if (type == M_MAIL && scan_file &&
		    (stat(itos(num), &s) < 0 ||
		     msginfo->size  != s.st_size ||
		     msginfo->mtime != s.st_mtime)) {
			procmsg_msginfo_free(msginfo);
		} else {
			msginfo->folder = g_strdup(folder);

			if (!mlist)
				last = mlist = g_slist_append(NULL, msginfo);
			else {
				last = g_slist_append(last, msginfo);
				last = last->next;
			}
		}
	}

	fclose(fp);
	debug_print(_("done.\n"));

	CHDIR_RETURN_VAL_IF_FAIL(maildir, mlist);
	return mlist;
}

static GSList *procmsg_get_uncached_msgs(GHashTable *msg_table,
					 const gchar *folder)
{
	DIR *dp;
	struct dirent *d;
	struct stat s;
	GSList *newlist = NULL;
	GSList *last = NULL;
	MsgInfo *msginfo;
	gint n_newmsg = 0;
	gint num;

	CHDIR_RETURN_VAL_IF_FAIL(maildir, NULL);
	CHDIR_RETURN_VAL_IF_FAIL(folder, NULL);

	if ((dp = opendir(".")) == NULL) {
		FILE_OP_ERROR(folder, "opendir");
		return NULL;
	}

	debug_print(_("\tSearching uncached messages... "));

	if (msg_table) {
		while ((d = readdir(dp)) != NULL) {
			if ((num = to_number(d->d_name)) < 0) continue;
			if (stat(d->d_name, &s) < 0) {
				FILE_OP_ERROR(d->d_name, "stat");
				continue;
			}
			if (!S_ISREG(s.st_mode)) continue;

			msginfo = g_hash_table_lookup
				(msg_table, GUINT_TO_POINTER(num));

			if (!msginfo) {
				/* not found in the cache (uncached message) */
				msginfo = procmsg_msginfo_set(d->d_name,
							      folder);
				if (!msginfo) continue;

				if (!newlist)
					last = newlist =
						g_slist_append(NULL, msginfo);
				else {
					last = g_slist_append(last, msginfo);
					last = last->next;
				}
				n_newmsg++;
			}
		}
	} else {
		/* discard all previous cache */
		while ((d = readdir(dp)) != NULL) {
			if (to_number(d->d_name) < 0) continue;
			if (stat(d->d_name, &s) < 0) {
				FILE_OP_ERROR(d->d_name, "stat");
				continue;
			}
			if (!S_ISREG(s.st_mode)) continue;

			msginfo = procmsg_msginfo_set(d->d_name, folder);
			if (!msginfo) continue;

			if (!newlist)
				last = newlist =
					g_slist_append(NULL, msginfo);
			else {
				last = g_slist_append(last, msginfo);
				last = last->next;
			}
			n_newmsg++;
		}
	}

	closedir(dp);

	if (n_newmsg)
		debug_print(_("%d uncached message(s) found.\n"), n_newmsg);
	else
		debug_print(_("done.\n"));

	/* sort new messages in numerical order */
	if (newlist) {
		debug_print(_("\tSorting uncached messages in numerical order... "));
		newlist = g_slist_sort
			(newlist, (GCompareFunc)procmsg_cmp_msgnum_for_sort);
		debug_print(_("done.\n"));
	}

	CHDIR_RETURN_VAL_IF_FAIL(maildir, newlist);
	return newlist;
}

void procmsg_set_flags(GSList *mlist, const gchar *folder, MsgType type)
{
	GSList *cur, *tmp;
	gint newmsg = 0;
	gchar *markdir;
	MsgInfo *msginfo;
	GHashTable *mark_table;
	MsgFlags flags;

	if (!mlist) return;
	g_return_if_fail(folder != NULL);

	debug_print(_("\tMarking the messages..."));

	markdir = g_strconcat(type == M_MAIL ? maildir : get_news_cache_dir(),
			      G_DIR_SEPARATOR_S, folder, NULL);
	mark_table = procmsg_read_mark_file(markdir);
	g_free(markdir);

	if (!mark_table) return;

	for (cur = mlist; cur != NULL; cur = cur->next) {
		msginfo = (MsgInfo *)cur->data;

		flags = GPOINTER_TO_UINT(g_hash_table_lookup(mark_table,
					 GUINT_TO_POINTER(msginfo->msgnum)));

		if (flags != 0) {
			/* flag found */
			/* preserve mime and cached flag */
			msginfo->flags &=
				(MSG_QUEUED|MSG_DRAFT|MSG_MIME|MSG_CACHED);
			msginfo->flags |=
				(flags &
				 ~(MSG_QUEUED|MSG_DRAFT|MSG_MIME|MSG_CACHED));
			if (type == M_NEWS)
				msginfo->flags |= MSG_NEWS;
		} else {
			/* not found (new message) */
			if (newmsg == 0) {
				for (tmp = mlist; tmp != cur; tmp = tmp->next)
					MSG_UNSET_FLAGS
						(((MsgInfo *)tmp->data)->flags,
						 MSG_NEW);
			}
			newmsg++;
		}
	}

	debug_print(_("done.\n"));
	if (newmsg)
		debug_print(_("\t%d new message(s)\n"), newmsg);

	g_hash_table_destroy(mark_table);
}

gint procmsg_get_last_num_in_cache(GSList *mlist)
{
	GSList *cur;
	MsgInfo *msginfo;
	gint last = 0;

	if (mlist == NULL) return 0;

	for (cur = mlist; cur != NULL; cur = cur->next) {
		msginfo = (MsgInfo *)cur->data;
		if (msginfo && msginfo->msgnum > last)
			last = msginfo->msgnum;
	}

	return last;
}

void procmsg_write_cache(MsgInfo *msginfo, FILE *fp)
{
	MsgFlags flags = msginfo->flags & MSG_MIME;

	WRITE_CACHE_DATA_INT(msginfo->msgnum, fp);
	WRITE_CACHE_DATA_INT(msginfo->size, fp);
	WRITE_CACHE_DATA_INT(msginfo->mtime, fp);
	WRITE_CACHE_DATA_INT(msginfo->date_t, fp);
	WRITE_CACHE_DATA_INT(flags, fp);

	WRITE_CACHE_DATA(msginfo->fromname, fp);

	WRITE_CACHE_DATA(msginfo->date, fp);
	WRITE_CACHE_DATA(msginfo->from, fp);
	WRITE_CACHE_DATA(msginfo->to, fp);
	WRITE_CACHE_DATA(msginfo->newsgroups, fp);
	WRITE_CACHE_DATA(msginfo->subject, fp);
	WRITE_CACHE_DATA(msginfo->msgid, fp);
	WRITE_CACHE_DATA(msginfo->inreplyto, fp);
	WRITE_CACHE_DATA(msginfo->default_mime_charset, fp);

}

void procmsg_write_flags(MsgInfo *msginfo, FILE *fp)
{
	MsgFlags flags;

	flags = msginfo->flags & (MSG_NEW|MSG_UNREAD|MSG_IMPORTANT|MSG_DELETED);
	//|MSG_MOVE|MSG_COPY

	WRITE_CACHE_DATA_INT(msginfo->msgnum, fp);
	WRITE_CACHE_DATA_INT(flags, fp);
}

struct MarkSum {
	gint *new;
	gint *unread;
	gint *total;
};

static void mark_sum_func(gpointer key, gpointer value, gpointer data)
{
	MsgFlags flags = GPOINTER_TO_UINT(value);
	struct MarkSum *marksum = data;

	if (MSG_IS_NEW(flags)) (*marksum->new)++;
	if (MSG_IS_UNREAD(flags)) (*marksum->unread)++;
	(*marksum->total)++;
}

void procmsg_get_mark_sum(const gchar *folder,
			  gint *new, gint *unread, gint *total)
{
	GHashTable *mark_table;
	struct MarkSum marksum;

	*new = *unread = *total = 0;
	marksum.new    = new;
	marksum.unread = unread;
	marksum.total  = total;

	mark_table = procmsg_read_mark_file(folder);

	if (mark_table) {
		g_hash_table_foreach(mark_table, mark_sum_func, &marksum);
		g_hash_table_destroy(mark_table);
	}
}

static GHashTable *procmsg_read_mark_file(const gchar *folder)
{
	FILE *fp;
	GHashTable *mark_table = NULL;
	gint num;
	MsgFlags flags;

	if ((fp = procmsg_open_mark_file(folder, FALSE)) == NULL)
		return NULL;

	mark_table = g_hash_table_new(NULL, g_direct_equal);

	while (fread(&num, sizeof(num), 1, fp) == 1) {
		if (fread(&flags, sizeof(flags), 1, fp) != 1) break;
		MSG_SET_FLAGS(flags, MSG_CACHED);

		g_hash_table_insert(mark_table,
				    GUINT_TO_POINTER(num),
				    GUINT_TO_POINTER(flags));
	}

	fclose(fp);
	return mark_table;
}

FILE *procmsg_open_mark_file(const gchar *folder, gboolean append)
{
	gchar *markfile;
	FILE *fp;
	gint ver;

	markfile = g_strconcat(folder, G_DIR_SEPARATOR_S, MARK_FILE, NULL);

	if ((fp = fopen(markfile, "r")) == NULL)
		debug_print(_("Mark file not found.\n"));
	else if (fread(&ver, sizeof(ver), 1, fp) != 1 || MARK_VERSION != ver) {
		debug_print(_("Mark version is different (%d != %d). "
			      "Discarding it.\n"), ver, MARK_VERSION);
		fclose(fp);
		fp = NULL;
	}

	/* read mode */
	if (append == FALSE) {
		g_free(markfile);
		return fp;
	}

	if (fp) {
		/* reopen with append mode */
		fclose(fp);
		if ((fp = fopen(markfile, "a")) == NULL)
			g_warning(_("Can't open mark file with append mode.\n"));
	} else {
		/* open with overwrite mode if mark file doesn't exist or
		   version is different */
		if ((fp = fopen(markfile, "w")) == NULL)
			g_warning(_("Can't open mark file with write mode.\n"));
		else {
			ver = MARK_VERSION;
			WRITE_CACHE_DATA_INT(ver, fp);
		}
	}

	g_free(markfile);
	return fp;
}

gint procmsg_get_last_message_number(const gchar *dir)
{
	gint max = 0, num;
	DIR *dp;
	struct dirent *d;
	struct stat s;

	CHDIR_RETURN_VAL_IF_FAIL(maildir, -1);
	CHDIR_RETURN_VAL_IF_FAIL(dir, -1);

	if ((dp = opendir(".")) == NULL) {
		FILE_OP_ERROR(dir, "opendir");
		CHDIR_RETURN_VAL_IF_FAIL(maildir, -1);
		return -1;
	}

	while ((d = readdir(dp)) != NULL) {
		if (stat(d->d_name, &s) < 0 ||
		    !S_ISREG(s.st_mode)     ||
		    (num = to_number(d->d_name)) < 0) continue;
		if (max < num)
			max = num;
	}

	closedir(dp);

	CHDIR_RETURN_VAL_IF_FAIL(maildir, -1);
	debug_print(_("Last number in dir %s = %d\n"), dir, max);
	return max;
}

/* copy a message into the destdir and return its number */
gint procmsg_copy_message(const gchar *destdir, const gchar *src)
{
	gint num;
	gchar *destfile;

	num = procmsg_get_last_message_number(destdir);
	if (num < 0)
		return -1;

	destfile = g_strdup_printf("%s%c%d",
				   destdir, G_DIR_SEPARATOR, num + 1);
	if (copy_file(src, destfile) < 0) {
		g_warning(_("can't copy message %s to %s\n"), src, destfile);
		g_free(destfile);
		return -1;
	}

	g_free(destfile);
	return num + 1;
}

void procmsg_move_messages(const gchar *from, GSList *mlist)
{
	GSList *cur, *movelist = NULL;
	MsgInfo *msginfo;
	gchar *to_folder = NULL;

	g_return_if_fail(from != NULL);
	if (!mlist) return;

	for (cur = mlist; cur != NULL; cur = cur->next) {
		msginfo = (MsgInfo *)cur->data;
		if (!to_folder) {
			to_folder = g_strdup(msginfo->to_folder);
			movelist = g_slist_append(movelist, msginfo);
		} else if (!strcmp(to_folder, msginfo->to_folder)) {
			movelist = g_slist_append(movelist, msginfo);
		} else {
			procmsg_move_messages_with_dest
				(to_folder, from, movelist);
			g_slist_free(movelist);
			movelist = NULL;
			g_free(to_folder);
			to_folder = g_strdup(msginfo->to_folder);
			movelist = g_slist_append(movelist, msginfo);
		}
	}

	if (movelist) {
		procmsg_move_messages_with_dest(to_folder, from, movelist);
		g_slist_free(movelist);
		g_free(to_folder);
	}
}

void procmsg_move_messages_with_dest(const gchar *to, const gchar *from,
				     GSList *mlist)
{
	GSList *cur;
	MsgInfo *msginfo, newmsginfo;
	guint new_num;
	gchar *to_dir;
	gchar *to_path;
	FILE *fp;

	g_return_if_fail(to    != NULL);
	g_return_if_fail(from  != NULL);
	g_return_if_fail(mlist != NULL);

	if (!path_cmp(to, from)) {
		g_warning(_("procmsg_move_messages_with_dest(): "
			    "src folder is identical to dest.\n"));
		return;
	}

	to_dir = g_strconcat(maildir, G_DIR_SEPARATOR_S, to, NULL);
	new_num = procmsg_get_last_message_number(to_dir) + 1;
	if (new_num < 1)
		return;

	CHDIR_RETURN_IF_FAIL(maildir);
	CHDIR_RETURN_IF_FAIL(from);

	if ((fp = procmsg_open_mark_file(to_dir, TRUE)) == NULL)
		g_warning(_("Can't open mark file.\n"));

	cur = mlist;
	while (cur != NULL) {
		msginfo = (MsgInfo *)cur->data;
		to_path = g_strdup_printf("%s%c%d", to_dir, G_DIR_SEPARATOR,
					  new_num);
		debug_print(_("Moving message %s/%d to %s/%d ...\n"),
			    from, msginfo->msgnum, to, new_num);
		if (rename(itos(msginfo->msgnum), to_path) < 0) {
			FILE_OP_ERROR(to_path, "rename");
			g_free(to_path);
			if (EISDIR == errno) {
				new_num++;
				continue;
			} else if(EXDEV == errno) {
		                if (copy_file(itos(msginfo->msgnum), to_path) < 0) {
			                FILE_OP_ERROR(to_path, "move file");
                                } else {
                                        if(unlink(itos(msginfo->msgnum))!=0) {
			                        FILE_OP_ERROR(to_path, "unlink");
                                        }
                                }
                        } else {
				break;
                        }
		} else if (fp) {
			newmsginfo.msgnum = new_num;
			if (!strcmp(to, OUTBOX_DIR) ||
			    !strcmp(to, QUEUE_DIR)  ||
			    !strcmp(to, DRAFT_DIR)  ||
			    !strcmp(to, TRASH_DIR))
				newmsginfo.flags = 0;
			else
				newmsginfo.flags = msginfo->flags;
			procmsg_write_flags(&newmsginfo, fp);
		}
		new_num++;
		g_free(to_path);
		cur = cur->next;
	}

	if (fp) fclose(fp);
	g_free(to_dir);
	CHDIR_RETURN_IF_FAIL(maildir);
}

void procmsg_copy_messages(const gchar *from, GSList *mlist)
{
	GSList *cur, *movelist = NULL;
	MsgInfo *msginfo;
	gchar *to_folder = NULL;

	g_return_if_fail(from != NULL);
	if (!mlist) return;

	for (cur = mlist; cur != NULL; cur = cur->next) {
		msginfo = (MsgInfo *)cur->data;
		if (!to_folder) {
			to_folder = g_strdup(msginfo->to_folder);
			movelist = g_slist_append(movelist, msginfo);
		} else if (!strcmp(to_folder, msginfo->to_folder)) {
			movelist = g_slist_append(movelist, msginfo);
		} else {
			procmsg_copy_messages_with_dest
				(to_folder, from, movelist);
			g_slist_free(movelist);
			movelist = NULL;
			g_free(to_folder);
			to_folder = g_strdup(msginfo->to_folder);
			movelist = g_slist_append(movelist, msginfo);
		}
	}

	if (movelist) {
		procmsg_copy_messages_with_dest(to_folder, from, movelist);
		g_slist_free(movelist);
		g_free(to_folder);
	}
}

void procmsg_copy_messages_with_dest(const gchar *to, const gchar *from,
				     GSList *mlist)
{
	GSList *cur;
	MsgInfo *msginfo, newmsginfo;
	guint new_num;
	gchar *to_dir;
	gchar *to_path;
	FILE *fp;

	g_return_if_fail(to    != NULL);
	g_return_if_fail(from  != NULL);
	g_return_if_fail(mlist != NULL);

	if (!path_cmp(to, from)) {
		g_warning(_("procmsg_copy_messages_with_dest(): "
			    "src folder is identical to dest.\n"));
		return;
	}

	to_dir = g_strconcat(maildir, G_DIR_SEPARATOR_S, to, NULL);
	new_num = procmsg_get_last_message_number(to_dir) + 1;
	if (new_num < 1)
		return;

	CHDIR_RETURN_IF_FAIL(maildir);
	CHDIR_RETURN_IF_FAIL(from);

	if ((fp = procmsg_open_mark_file(to_dir, TRUE)) == NULL)
		g_warning(_("Can't open mark file.\n"));

	cur = mlist;
	while (cur != NULL) {
		msginfo = (MsgInfo *)cur->data;
		to_path = g_strdup_printf("%s%c%d", to_dir, G_DIR_SEPARATOR,
					  new_num);
		debug_print(_("Copying message %s/%d to %s/%d ...\n"),
			    from, msginfo->msgnum, to, new_num);
		if (copy_file(itos(msginfo->msgnum), to_path) < 0) {
			FILE_OP_ERROR(to_path, "copy file");
		} else if (fp) {
			newmsginfo.msgnum = new_num;
			if (!strcmp(to, OUTBOX_DIR) ||
			    !strcmp(to, QUEUE_DIR)  ||
			    !strcmp(to, DRAFT_DIR)  ||
			    !strcmp(to, TRASH_DIR))
				newmsginfo.flags = 0;
			else
				newmsginfo.flags = msginfo->flags;
			procmsg_write_flags(&newmsginfo, fp);
		}
		new_num++;
		g_free(to_path);
		cur = cur->next;
	}

	if (fp) fclose(fp);
	g_free(to_dir);
	CHDIR_RETURN_IF_FAIL(maildir);
}

FILE *procmsg_open_message(MsgInfo *msginfo)
{
	FILE *fp;
	gchar *file;

	g_return_val_if_fail(msginfo != NULL, NULL);

	file = (char*) procmsg_get_message_file_path(msginfo);
	g_return_val_if_fail(file != NULL, NULL);

	if ((fp = fopen(file, "r")) == NULL) {
		FILE_OP_ERROR(file, "fopen");
		g_free(file);
		return NULL;
	}

	g_free(file);

	if (MSG_IS_QUEUED(msginfo->flags)) {
		gchar buf[BUFFSIZE];

		while (fgets(buf, sizeof(buf), fp) != NULL)
			if (buf[0] == '\r' || buf[0] == '\n') break;
	}

	return fp;
}

gboolean procmsg_msg_exist(MsgInfo *msginfo)
{
	gchar *path;
	struct stat s;

	if (!msginfo) return FALSE;

	path = (gchar*) procmsg_get_message_file_path(msginfo);
	if (stat(path, &s) == 0 &&
	    msginfo->size  == s.st_size &&
	    msginfo->mtime == s.st_mtime)
		return TRUE;
	else
		return FALSE;
}

void procmsg_clean_trash(void)
{
	DIR *dp;
	struct dirent *d;

	CHDIR_RETURN_IF_FAIL(maildir);
	CHDIR_RETURN_IF_FAIL(TRASH_DIR);

	if ((dp = opendir(".")) == NULL) {
		FILE_OP_ERROR(TRASH_DIR, "opendir");
		CHDIR_RETURN_IF_FAIL(maildir);
		return;
	}

	while ((d = readdir(dp)) != NULL) {
		if (to_number(d->d_name) < 0) continue;
		debug_print(_("Unlinking message %s in trash...\n"), d->d_name);
		if (unlink(d->d_name) < 0)
			FILE_OP_ERROR(d->d_name, "unlink");
	}

	debug_print(_("done.\n"));
	closedir(dp);

	CHDIR_RETURN_IF_FAIL(maildir);
}

gint procmsg_send_queue(void)
{
	gint num;
	gint i;
	gchar *msg;

	num = procmsg_get_last_message_number(QUEUE_DIR);
	if (num < 0) return -1;
	else if (num == 0) return 0;

	CHDIR_RETURN_VAL_IF_FAIL(QUEUE_DIR, -1);

	for (i = 1; i <= num; i++) {
		msg = itos(i);
		if (is_file_exist(msg)) {
			if (send_message_queue(msg) < 0) {
				g_warning(_("Sending queued message failed.\n"));
				change_dir(maildir);
				return -1;
			}
		}
	}

	change_dir(maildir);
	return 0;
}

void procmsg_print_message(MsgInfo *msginfo, const gchar *cmdline)
{
	static const gchar *def_cmd = "lpr %s";
	static guint id = 0;
	gchar *prtmp;
	FILE *tmpfp, *prfp;
	gchar buf[1024];
	gchar *p;

	g_return_if_fail(msginfo);
	if ((tmpfp = procmime_get_text_part(msginfo)) == NULL) {
		g_warning(_("Can't get text part\n"));
		return;
	}
	prtmp = g_strdup_printf("%s%cprinttmp.%08x",
				get_mime_tmp_dir(), G_DIR_SEPARATOR, id++);

	if ((prfp = fopen(prtmp, "w")) == NULL) {
		FILE_OP_ERROR(prtmp, "fopen");
		g_free(prtmp);
		fclose(tmpfp);
		return;
	}

	if (msginfo->date) fprintf(prfp, "Date: %s\n", msginfo->date);
	if (msginfo->from) fprintf(prfp, "From: %s\n", msginfo->from);
	if (msginfo->to)   fprintf(prfp, "To: %s\n", msginfo->to);
	if (msginfo->newsgroups)
		fprintf(prfp, "Newsgroups: %s\n", msginfo->newsgroups);
	if (msginfo->subject) fprintf(prfp, "Subject: %s\n", msginfo->subject);
	fputc('\n', prfp);

	while (fgets(buf, sizeof(buf), tmpfp) != NULL)
		fputs(buf, prfp);

	fclose(prfp);
	fclose(tmpfp);

	if (cmdline && (p = strchr(cmdline, '%')) && *(p + 1) == 's' &&
	    !strchr(p + 2, '%'))
		g_snprintf(buf, sizeof(buf) - 1, cmdline, prtmp);
	else {
		if (cmdline)
			g_warning(_("Print command line is invalid: `%s'\n"),
				  cmdline);
		g_snprintf(buf, sizeof(buf) - 1, def_cmd, prtmp);
	}

	g_free(prtmp);

	g_strchomp(buf);
	if (buf[strlen(buf) - 1] != '&') strcat(buf, "&");
	system(buf);
}

MsgInfo *procmsg_msginfo_set(const gchar *msgnum, const gchar *folder)
{
	gchar *file;
	struct stat s;
	MsgInfo *msginfo;

	MsgFlags flags = MSG_NEW|MSG_UNREAD;

	g_return_val_if_fail(folder != NULL, NULL);
	g_return_val_if_fail(msgnum != NULL, NULL);

	file = g_strconcat(maildir, G_DIR_SEPARATOR_S,
			   folder, G_DIR_SEPARATOR_S, msgnum, NULL);
	if (!strcmp(folder, QUEUE_DIR)) {
		MSG_SET_FLAGS(flags, MSG_QUEUED);
	} else if (!strcmp(folder, DRAFT_DIR)) {
		MSG_SET_FLAGS(flags, MSG_DRAFT);
	}

	file = g_strdup_printf("%s/%s/%s", maildir, folder, msgnum);
	msginfo = procheader_parse(file, flags, FALSE);
	if (!msginfo) {
		g_free(file);
		return NULL;
	}

	msginfo->msgnum = atoi(msgnum);
	msginfo->folder = g_strdup(folder);

	if (stat(file, &s) < 0) {
		FILE_OP_ERROR(file, "stat");
		msginfo->size = 0;
		msginfo->mtime = 0;
	} else {
		msginfo->size = s.st_size;
		msginfo->mtime = s.st_mtime;
	}

	g_free(file);

	return msginfo;
}

MsgInfo *procmsg_msginfo_copy(MsgInfo *msginfo)
{
	MsgInfo *newmsginfo;

	if (msginfo == NULL) return NULL;
	newmsginfo = g_new0(MsgInfo, 1);

#define MEMBCOPY(mmb)	newmsginfo->mmb = msginfo->mmb
#define MEMBDUP(mmb)	newmsginfo->mmb = msginfo->mmb ? \
			g_strdup(msginfo->mmb) : NULL

	MEMBCOPY(msgnum);
	MEMBCOPY(size);
	MEMBCOPY(mtime);
	MEMBCOPY(date_t);
	MEMBCOPY(flags);

	MEMBDUP(fromname);

	MEMBDUP(date);
	MEMBDUP(from);
	MEMBDUP(to);
	MEMBDUP(newsgroups);
	MEMBDUP(subject);
	MEMBDUP(msgid);
	MEMBDUP(inreplyto);
	MEMBDUP(default_mime_charset);

	MEMBDUP(folder);
	MEMBDUP(to_folder);

	MEMBDUP(xface);

	return newmsginfo;
}

void procmsg_msginfo_free(MsgInfo *msginfo)
{
	if (msginfo == NULL) return;

	g_free(msginfo->folder);
	g_free(msginfo->to_folder);
	g_free(msginfo->xface);
	g_free(msginfo->fromname);

	g_free(msginfo->date);
	g_free(msginfo->from);
	g_free(msginfo->to);
	g_free(msginfo->newsgroups);
	g_free(msginfo->subject);
	g_free(msginfo->msgid);
	g_free(msginfo->inreplyto);
	g_free(msginfo->default_mime_charset);

	g_free(msginfo);
}

static gint procmsg_cmp_msgnum(gconstpointer a, gconstpointer b)
{
	const MsgInfo *msginfo = a;
	const guint msgnum = GPOINTER_TO_UINT(b);

	if (!msginfo)
		return -1;

	return msginfo->msgnum - msgnum;
}

static gint procmsg_cmp_msgnum_for_sort(gconstpointer a, gconstpointer b)
{
	const MsgInfo *msginfo1 = a;
	const MsgInfo *msginfo2 = b;

	if (!msginfo1)
		return -1;
	if (!msginfo2)
		return -1;

	return msginfo1->msgnum - msginfo2->msgnum;
}

static gint procmsg_cmp_flag_msgnum(gconstpointer a, gconstpointer b)
{
	const FlagInfo *finfo = a;
	const guint msgnum = GPOINTER_TO_UINT(b);

	if (!finfo)
		return -1;

	return finfo->msgnum - msgnum;
}

gchar *procmsg_get_message_file_path(MsgInfo *msginfo)
{
	g_return_val_if_fail(msginfo != NULL, NULL);

	return g_strconcat(MSG_IS_NEWS(msginfo->flags) ?
			   get_news_cache_dir() : maildir,
			   G_DIR_SEPARATOR_S, msginfo->folder,
			   G_DIR_SEPARATOR_S, itos(msginfo->msgnum), NULL);
}

gchar *procmsg_get_message_file(MsgInfo *msginfo, gpointer data)
{
	gint ok;
	gchar *filename;

	g_return_val_if_fail(msginfo != NULL, NULL);

	if (MSG_IS_NEWS(msginfo->flags)) {
		NNTPSession *session = (NNTPSession *)data;

		filename = g_strconcat(get_news_cache_dir(), G_DIR_SEPARATOR_S,
				       msginfo->folder, G_DIR_SEPARATOR_S,
				       itos(msginfo->msgnum), NULL);
		if (!is_file_exist(filename)) {
			if (!session) {
				g_warning(_("news session is not established\n"));
				g_free(filename);
				return NULL;
			}

			debug_print(_("getting article %d...\n"),
				    msginfo->msgnum);
			ok = news_get_article
				(session, msginfo->msgnum, filename);
			statusbar_pop_all();
			if (ok < 0) {
				g_warning(_("can't read article %d\n"),
					  msginfo->msgnum);
				g_free(filename);
				return NULL;
			}
			debug_print(_("done.\n"));
		} else
			debug_print(_("article %d has been already cached.\n"),
				    msginfo->msgnum);
	} else {
		filename = g_strconcat(maildir, G_DIR_SEPARATOR_S,
				       msginfo->folder, G_DIR_SEPARATOR_S,
				       itos(msginfo->msgnum), NULL);
	}

	return filename;
}

