/*
    Stuphead: (C) 2000,2001 Grigroy Bakunov, Sergey Pinaev
 */
/*
 * Sylpheed -- a GTK+ based, lightweight, and fast e-mail client
 * Copyright (C) 1999,2000 Hiroyuki Yamamoto
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <glib.h>
#include <stdio.h>
#include <string.h>

#include "intl.h"
#include "main.h"
#include "send.h"
#include "socket.h"
#include "smtp.h"
#include "nntp.h"
#include "account.h"
#include "compose.h"
#include "procmsg.h"
#include "procheader.h"
#include "utils.h"

static gint send_message_smtp	(GSList *to_list, const gchar *from,
				 const gchar *server, FILE *fp);
static gint send_smtp_open	(void);

static gint send_message_nntp	(GSList *to_list, const gchar *from,
				 const gchar *server, FILE *fp);
static gint send_nntp_open	(const gchar* server);

#define SEND_EXIT_IF_ERROR(f, s) \
{ \
	if ((f) < 0) { \
		log_warning("Error occured while %s\n", s); \
		if (smtp_sock > 0) sock_close(smtp_sock); \
		return -1; \
	} \
}

#define SEND_NG_EXIT_IF_ERROR(f, s) \
{ \
	if ((f) < 0) { \
		log_warning("Error occured while %s\n", s); \
		if (nntp_sock > 0) sock_close(nntp_sock); \
		return -1; \
	} \
}

#define SEND_EXIT_IF_NOTOK(f, s) \
{ \
	if ((f) != SM_OK) { \
		log_warning("Error occured while %s\n", s); \
		if (smtp_quit(smtp_sock) != SM_OK) \
			log_warning("Error occured while sending QUIT\n"); \
		sock_close(smtp_sock); \
		return -1; \
	} \
}

#define SEND_NG_EXIT_IF_NOTOK(f, s) \
{ \
	if ((f) != NN_SUCCESS) { \
		log_warning("Error occured while %s\n", s); \
		if (nntp_quit(nntp_sock) != NN_SUCCESS) \
			log_warning("Error occured while sending QUIT\n"); \
		sock_close(nntp_sock); \
		return -1; \
	} \
}
gint send_message	(GSList *to_list,    GSList *ng_list, 
			 const gchar *nserver,
			 const gchar *files, const gchar *filen,
			 gboolean ismail,    gboolean isnews)
{
	FILE *fp;
	gint val1, val2;
	val1=val2=0;
	g_return_val_if_fail(files != NULL, -1);
	g_return_val_if_fail(filen != NULL, -1);	
	g_return_val_if_fail(cur_account != NULL, -1);
	if (ismail) {
		if ((fp = fopen(files, "r")) == NULL) {
		    FILE_OP_ERROR(files, "fopen");
		    return -1;
		};
		val1 = send_message_smtp(to_list, cur_account->address, cur_account->smtp_server, fp);
		fclose(fp);
	};

	if (isnews) {
		if ((fp = fopen(filen, "r")) == NULL) {
		    FILE_OP_ERROR(filen, "fopen");
		    return -1;
		};
	
		val2 = send_message_nntp(ng_list, cur_account->address, nserver, fp);
		fclose(fp);
	};
	if (val1<0 || val2<0) return -1;
	return val1+val2;
}

enum
{
	Q_SENDER     = 0,
	Q_SMTPSERVER = 1,
	Q_RECIPIENTS = 2,
	Q_NNTPSERVER = 3,
	Q_NEWSGROUPS = 4,
	
};

gint send_message_queue(const gchar *file)
{
	static HeaderEntry qentry[] = {{"S:",   NULL, FALSE},
				       {"SSV:", NULL, FALSE},
				       {"R:",   NULL, FALSE},
				       {"NSV:", NULL, FALSE},
				       {"N:"  , NULL, FALSE},				       
				       {NULL,   NULL, FALSE}};
	FILE *fp;
	gint vals=0;
	gint valn=0;
	gint val=0;
	gchar *from = NULL;
	gchar *sserver = NULL;
	gchar *nserver = NULL;	
	GSList *to_list = NULL;
	GSList *ng_list = NULL;	
	gchar buf[BUFFSIZE];
	gint hnum;
	long pos;

	g_return_val_if_fail(file != NULL, -1);

	if ((fp = fopen(file, "r")) == NULL) {
		FILE_OP_ERROR(file, "fopen");
		return -1;
	}

	while ((hnum = procheader_get_one_field(buf, sizeof(buf), fp, qentry))
	       != -1) {
		gchar *p = buf + strlen(qentry[hnum].name);

		switch (hnum) {
		case Q_SENDER:
			from = g_strdup(p);
			break;
		case Q_SMTPSERVER:
			sserver = g_strdup(p);
			break;
		case Q_RECIPIENTS:
			to_list = address_list_append(to_list, p);
			break;
		case Q_NNTPSERVER:
			nserver = g_strdup(p);
			break;
		case Q_NEWSGROUPS:
			ng_list = address_list_append(ng_list, p);
			break;

		default:
		}
	}

	if (!from) {
		g_warning(_("Queued message header is broken.\n"));
		val = -1;
	} else {
		pos=ftell(fp);
		if (pos !=-1) {
    		    fseek(fp, pos, SEEK_SET);
    		    if (sserver!=NULL && to_list!=NULL) {
    			vals = send_message_smtp(to_list, from, sserver, fp);
			slist_free_strings(to_list);
			g_slist_free(to_list);
			g_free(sserver);

		    };
    		    fseek(fp, pos, SEEK_SET);
		    if (nserver!=NULL && ng_list!=NULL) {
			valn = send_message_nntp(ng_list, from, nserver, fp);
			slist_free_strings(ng_list);
			g_slist_free(ng_list);
			g_free(nserver);
		    };

		    if (vals !=0 || valn !=0) val=-2;
		    if (val == 0) {
		    	    if (unlink(file) < 0)
			    	    FILE_OP_ERROR(file, "unlink");
		    };
		} else {
			g_warning(_("Error in sending queue process.\n"));
		};

	    g_free(from);
	}
	fclose(fp);
	return val;
}

static gint send_message_smtp(GSList *to_list, const gchar *from,
			      const gchar *server, FILE *fp)
{
	gint smtp_sock;
	gchar buf[BUFFSIZE];
	const gchar *smtp_userid;
	const gchar *smtp_passwd;
	GSList *cur;

	g_return_val_if_fail(to_list != NULL, -1);
	g_return_val_if_fail(from != NULL, -1);
	g_return_val_if_fail(server != NULL, -1);
	g_return_val_if_fail(fp != NULL, -1);
        g_return_val_if_fail(cur_account != NULL, -1);

	log_message(_("Connecting to SMTP server: %s ...\n"), server);

	SEND_EXIT_IF_ERROR((smtp_sock = send_smtp_open()),
			   "connecting to server");

	if (cur_account->smtp_auth) {
		smtp_userid = cur_account->smtp_userid;
		smtp_passwd = cur_account->smtp_passwd;

		switch (cur_account->a_protocol) {
		    case A_LOGIN: 
				SEND_EXIT_IF_NOTOK(smtp_auth_challenge(smtp_sock, A_LOGIN), "sending AUTH LOGIN");
				SEND_EXIT_IF_NOTOK(smtp_auth_login(smtp_sock, smtp_userid, smtp_passwd), "sending AUTH data");
				break;
		    case A_DIGMD5:
				SEND_EXIT_IF_NOTOK(smtp_auth_challenge(smtp_sock, A_DIGMD5), "sending AUTH DIGEST-MD5");
				SEND_EXIT_IF_NOTOK(smtp_auth_digmd5(smtp_sock, smtp_userid, smtp_passwd), "sending AUTH data");
				break;
		    case A_CRMMD5:
				SEND_EXIT_IF_NOTOK(smtp_auth_challenge(smtp_sock, A_CRMMD5), "sending AUTH CRAM-MD5");
				SEND_EXIT_IF_NOTOK(smtp_auth_crmmd5(smtp_sock, smtp_userid, smtp_passwd), "sending AUTH data");
				break;
		}
	}

	SEND_EXIT_IF_NOTOK(smtp_from(smtp_sock, from), "sending MAIL FROM");
	for (cur = to_list; cur != NULL; cur = cur->next)
		SEND_EXIT_IF_NOTOK(smtp_rcpt(smtp_sock, (gchar *)cur->data),
				   "sending RCPT TO");
	SEND_EXIT_IF_NOTOK(smtp_data(smtp_sock), "sending DATA");

	/* send main part */
	while (fgets(buf, sizeof(buf), fp) != NULL) {
		strretchomp(buf);

		/* escape when a dot appears on the top */
		if (buf[0] == '.')
			SEND_EXIT_IF_ERROR(sock_write(smtp_sock, ".", 1),
					   "sending data");

		SEND_EXIT_IF_ERROR(sock_puts(smtp_sock, buf), "sending data");
	}

	SEND_EXIT_IF_NOTOK(smtp_eom(smtp_sock), "terminating data");
	SEND_EXIT_IF_NOTOK(smtp_quit(smtp_sock), "sending QUIT");

	sock_close(smtp_sock);
	return 0;
}

static gint send_smtp_open(void)
{
	SockInfo *sockinfo;
	gint sock;
	const gchar *server;
	gushort port;
	gint val;

	g_return_val_if_fail(cur_account != NULL, -1);

	server = cur_account->smtp_server;
	port = cur_account->set_smtpport ? cur_account->smtpport : 25;

	if ((sockinfo = sock_connect(server, port)) == NULL) {
		log_warning(_("Can't connect to SMTP server: %s:%d\n"),
			    server, port);
		return -1;
	}
	sock = sockinfo->sock;
	sock_sockinfo_free(sockinfo);
	if (smtp_ok(sock) == SM_OK) {
		if (cur_account->set_domain &&
		    cur_account->domain && *cur_account->domain)
		    {
			if (cur_account->smtp_auth) 
			    val = smtp_ehlo(sock, cur_account->domain); 
			else
			    val = smtp_helo(sock, cur_account->domain);
		    }
		else {

			if (cur_account->smtp_auth) 
			    val = smtp_ehlo(sock, get_domain_name());
			else
			    val = smtp_helo(sock, get_domain_name());
		    };
		if (val == SM_OK) return sock;
	}

	log_warning(_("Error occured while connecting to server\n"));
	sock_close(sock);

	return -1;
}

static gint send_nntp_open(const gchar* server)
{
	SockInfo *sockinfo;
	gint sock;
	gushort port;
	gint val;

	g_return_val_if_fail(cur_account != NULL, -1);
        if (server==NULL) {
            debug_print(_("Whats up ? No NNTP server settings!\n"));
            return -1;
        };
	port = cur_account->set_nntpport ? cur_account->nntpport : 119;

	if ((sockinfo = sock_connect(server, port)) == NULL) {
		log_warning(_("Can't connect to NNTP server: %s:%d\n"),
			    server, port);
		return -1;
	}
	sock = sockinfo->sock;
	sock_sockinfo_free(sockinfo);
	if (nntp_ok(sock, NULL) == NN_SUCCESS) return sock;
	log_warning(_("Error occured while connecting to server\n"));
	sock_close(sock);

	return -1;
}


static gint send_message_nntp(GSList *to_list, const gchar *from,
			      const gchar *server, FILE *fp)
{
	gint nntp_sock;
	gchar buf[BUFFSIZE];
	GSList *cur;

	g_return_val_if_fail(to_list != NULL, -1);
	g_return_val_if_fail(from != NULL, -1);
	g_return_val_if_fail(fp != NULL, -1);
        g_return_val_if_fail(cur_account != NULL, -1);
	
	if (server==NULL) 
	    if (cur_account->nntp_server!=NULL) server=g_strdup(cur_account->nntp_server);
	    else return 0;
	log_message(_("Connecting to NNTP server: %s ...\n"), server);

	SEND_NG_EXIT_IF_ERROR((nntp_sock = send_nntp_open(server)),	   "connecting to server");
	if (cur_account->nntp_auth) {
		log_message(_("Trying to set AUTHINFO.\n"));
		if (nntp_auth(nntp_sock,cur_account->nntp_userid,cur_account->nntp_passwd) != NN_SUCCESS) {
			alertpanel_error(_("Cannot connect authorize with server."));
			return -1;
		};
	};
	SEND_NG_EXIT_IF_NOTOK(nntp_mode(nntp_sock, FALSE), "set MODE READER");
	SEND_NG_EXIT_IF_NOTOK(nntp_post(nntp_sock), "set POST mode");

	while (fgets(buf, sizeof(buf), fp) != NULL) {
		strretchomp(buf);
		/* escape when a dot appears on the top */
		if (buf[0] == '.')
			SEND_NG_EXIT_IF_ERROR(sock_write(nntp_sock, ".", 1),
					   "sending data");

		SEND_NG_EXIT_IF_ERROR(sock_puts(nntp_sock, buf), "sending data");
	}

	SEND_NG_EXIT_IF_NOTOK(nntp_eom(nntp_sock), "terminating data");

	sock_close(nntp_sock);
	return 0;
}
