/*
    Stuphead: (C) 2000,2001 Grigroy Bakunov, Sergey Pinaev
 */
/*
 * Sylpheed -- a GTK+ based, lightweight, and fast e-mail client
 * Copyright (C) 1999,2000 Hiroyuki Yamamoto
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <glib.h>
#include <gdk/gdk.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtkscrolledwindow.h>
#include <gtk/gtktext.h>
#include <gtk/gtksignal.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>

#include "main.h"
#include "summaryview.h"
#include "procheader.h"
#include "prefs_common.h"
#include "utils.h"
#include "gtkutils.h"
#include "procmime.h"

typedef struct _RemoteURI	RemoteURI;

struct _RemoteURI
{
	gchar *uri;

	guint start;
	guint end;
};


static GdkFont *msgfont;
static GdkFont *normalfont;
static GdkFont *boldfont;
static GdkFont *spacingfont;

static void textview_write_line		(TextView	*textview,
					 const gchar	*str,
					 gint 	ict);

static void textview_scan_header	(TextView	*textview,
					 FILE		*fp,
					 const gchar *charset);

static void textview_key_pressed	(GtkWidget	*widget,
					 GdkEventKey	*event,
					 TextView	*textview);
static void textview_button_pressed	(GtkWidget	*widget,
					 GdkEventButton	*event,
					 TextView	*textview);

static void textview_uri_list_remove_all(GSList		*uri_list);

TextView *textview_create(void)
{
	TextView *textview;
	GtkWidget *scrolledwin;
	GtkWidget *text;

	debug_print(_("Creating text view...\n"));
	textview = g_new0(TextView, 1);

	scrolledwin = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolledwin),
				       GTK_POLICY_NEVER, GTK_POLICY_ALWAYS);

	text = gtk_text_new(gtk_scrolled_window_get_hadjustment
			    (GTK_SCROLLED_WINDOW(scrolledwin)),
			    gtk_scrolled_window_get_vadjustment
			    (GTK_SCROLLED_WINDOW(scrolledwin)));
	gtk_widget_set_usize(scrolledwin, prefs_common.mainview_width, -1);
	gtk_container_add(GTK_CONTAINER(scrolledwin), text);
	gtk_signal_connect(GTK_OBJECT(text), "key_press_event",
			   GTK_SIGNAL_FUNC(textview_key_pressed),
			   textview);
	gtk_signal_connect(GTK_OBJECT(text), "button_press_event",
			   GTK_SIGNAL_FUNC(textview_button_pressed),
			   textview);

	gtk_widget_show_all(scrolledwin);
	gtk_text_set_word_wrap(GTK_TEXT(text), TRUE);


	textview->scrolledwin = scrolledwin;
	textview->text        = text;
	textview->uri_list    = NULL;

	return textview;
}

void textview_init(TextView *textview)
{
	GdkColormap *colormap;
	GdkColor color[3];
	gboolean success[3];

	gtkut_widget_disable_theme_engine(textview->text);
	color[0] = prefs_common.quote_color;
	color[1] = prefs_common.uri_color;
	color[2] = prefs_common.emphasis_color;

	colormap = gdk_window_get_colormap(textview->text->window);
	gdk_colormap_alloc_colors(colormap, color, 3, FALSE, TRUE, success);
	if (success[0] == FALSE || success[1] == FALSE ||
	    success[2] == FALSE) {
		GtkStyle *style;

		g_warning(_("TextView: color allocation failed\n"));
		style = gtk_widget_get_style(textview->text);
		prefs_common.quote_color = prefs_common.uri_color = prefs_common.emphasis_color = style->black;
	}

	textview_set_font();
}

void textview_show_message(TextView *textview, MimeInfo *mimeinfo,
			   const gchar *file)
{
	FILE *fp;

	if ((fp = fopen(file, "r")) == NULL) {
		FILE_OP_ERROR(file, "fopen");
		return;
	}

	textview_show_part(textview, mimeinfo, fp);

	fclose(fp);
}

void textview_show_part(TextView *textview, MimeInfo *mimeinfo, FILE *fp)
{
	GtkText *text = GTK_TEXT(textview->text);
	gchar buf[BUFFSIZE];
	const gchar *boundary = NULL;
	gint boundary_len = 0;
	gchar *charset = NULL;
	FILE *tmpfp;
	int len;

	g_return_if_fail(mimeinfo != NULL);
	g_return_if_fail(fp != NULL);

	textview_clear(textview);

	if (prefs_common.force_charset) {
		charset = g_strdup(prefs_common.force_charset);
		debug_print("changed to fc2 %s\n",charset);
		}
	else if (mimeinfo->charset) {
		charset = (gchar*) get_iconv_charset(mimeinfo->charset);
		debug_print("changed to mi %s from %s\n",charset, mimeinfo->charset);
		}
	else charset = (gchar*)get_iconv_charset(prefs_common.default_mime_charset);

	if (mimeinfo->mime_type == MIME_MULTIPART) {
		if (mimeinfo->sub) {
			mimeinfo = mimeinfo->sub;
			if (fseek(fp, mimeinfo->fpos, SEEK_SET) < 0) {
				perror("fseek");
				return;
			}
		} else
			return;
	}
	if (mimeinfo->parent && mimeinfo->parent->boundary) {
		boundary = mimeinfo->parent->boundary;
		boundary_len = strlen(boundary);
	}

	gtk_text_freeze(text);

	if (!boundary && mimeinfo->mime_type == MIME_TEXT) {
		if (fseek(fp, mimeinfo->fpos, SEEK_SET) < 0)
			perror("fseek");
		textview_scan_header(textview, fp, charset);
	} else {
		if (mimeinfo->mime_type == MIME_TEXT) {
			glong fpos;

			if ((fpos = ftell(fp)) < 0)
				perror("ftell");
 			else if (fseek(fp, mimeinfo->parent->fpos, SEEK_SET)
 				 < 0)
				perror("fseek");
			else {
				textview_scan_header(textview, fp, charset);
				if (fseek(fp, fpos, SEEK_SET) < 0)
					perror("fseek");
			}
		}
		while (fgets(buf, sizeof(buf), fp) != NULL)
			if (buf[0] == '\r' || buf[0] == '\n') break;
	}

	if (mimeinfo->parent && mimeinfo->mime_type == MIME_MESSAGE_RFC822) {
		textview_scan_header(textview, fp, charset);
		mimeinfo = mimeinfo->sub;
		if (!mimeinfo) {
			gtk_text_thaw(text);
			return;
		}
	}

	g_free(charset);
	if (prefs_common.force_charset) {
		charset = g_strdup(prefs_common.force_charset);
		debug_print("changed to fc2 %s\n",charset);
		}
	else if (mimeinfo->charset) {
		charset = (gchar*) get_iconv_charset(mimeinfo->charset);
		debug_print("changed to mi %s from %s\n",charset, mimeinfo->charset);
		}
	else charset = (gchar*)get_iconv_charset(prefs_common.default_mime_charset);

	tmpfp = procmime_decode_content(NULL, fp, mimeinfo);
	if (tmpfp) {
//		if (mimeinfo->mime_type == MIME_TEXT_HTML)
//			textview_show_html(textview, tmpfp, charset);
//		else
		gint ict=conv_iconv_open_tolocal(charset);
		while (fgets(buf, sizeof(buf), tmpfp) != NULL) {
			len = strlen(buf);
			if (buf[len-1] == '\n' && buf[len-2] == '\r') {
			    buf[len-2] = '\n';
			    buf[len-1] = '\0';
			};
			textview_write_line(textview, buf, ict);
		};
		conv_iconv_close(ict);
		fclose(tmpfp);
	}
	g_free(charset);

	gtk_text_thaw(text);
}

void textview_show_mime_part(TextView *textview, MimeInfo *partinfo)
{
	GtkText *text = GTK_TEXT(textview->text);

	if (!partinfo) return;

	textview_clear(textview);

	gtk_text_freeze(text);

	gtk_text_insert
		(text, msgfont, NULL, NULL,
		 _("To save this part, pop up the context menu with\n"), -1);
	gtk_text_insert
		(text, msgfont, NULL, NULL,
		 _("right click and select `Save as...', or press `y' key.\n\n"), -1);

	gtk_text_insert
		(text, msgfont, NULL, NULL,
		 _("To display this part as a text message, select\n"), -1);
	gtk_text_insert
		(text, msgfont, NULL, NULL,
		 _("`Display as text', or press `t' key.\n\n"), -1);

	gtk_text_insert
		(text, msgfont, NULL, NULL,
		 _("To open this part with external program, select `Open',\n"), -1);
	gtk_text_insert
		(text, msgfont, NULL, NULL,
		 _("or double-click, or click the center button, or press `l' key."), -1);

	gtk_text_thaw(text);
}

static void textview_write_line(TextView *textview, const gchar *str,
				gint ict)
{
	gchar buf[BUFFSIZE];
	gchar *p;
	GdkColor *fg_color;
	gint retval;
	GtkText *text;
	if (textview==NULL ||textview->text==NULL || str==NULL)   return;
	text = GTK_TEXT(textview->text);
	retval=conv_iconv_convert(ict,buf, BUFFSIZE,   str, strlen(str)+1);
	//if (retval<0) return;
	if (prefs_common.conv_mb_alnum) conv_mb_alnum(buf);
	fg_color = NULL;
	/* change color of quotation */
	/* >, foo>, _> ... ok, <foo>, foo bar>, foo-> ... ng */
	if (prefs_common.enable_color && ((void*) get_quotemark_position(buf) != NULL))
	  fg_color = &prefs_common.quote_color;

	if (prefs_common.head_space && spacingfont && buf[0] != '\n')
		gtk_text_insert(text, spacingfont, NULL, NULL, " ", 1);

	/* parse URI */
	if (prefs_common.enable_color &&
	    ((p = strcasestr(buf, "http://"))  != NULL ||
	     (p = strcasestr(buf, "https://")) != NULL ||
	     (p = strcasestr(buf, "ftp://"))   != NULL)) {
		gchar *ep;
		RemoteURI *uri;

		uri = g_new(RemoteURI, 1);
                gtk_text_insert(text, msgfont, fg_color, NULL, buf, p - buf);
		/* find end point of URI */
		for (ep = p; *ep; ep++)
			if (!isgraph(*ep) || !isascii(*ep) ||
			    strchr("()<>\"", *ep))
				break;

		uri->uri = g_strndup(p, ep - p);
		uri->start = gtk_text_get_point(text);
                gtk_text_insert(text, msgfont, &prefs_common.uri_color, NULL, p, ep - p);
	       	uri->end = gtk_text_get_point(text);

		textview->uri_list = g_slist_append(textview->uri_list, uri);

		gtk_text_insert(text, msgfont, fg_color, NULL, ep, -1);
	} else	gtk_text_insert(text, msgfont, fg_color, NULL, buf, -1);
}

void textview_clear(TextView *textview)
{
	GtkText *text = GTK_TEXT(textview->text);

	gtk_text_freeze(text);
	gtk_text_backward_delete(text, gtk_text_get_length(text));
	gtk_text_thaw(text);

	textview_uri_list_remove_all(textview->uri_list);
	textview->uri_list = NULL;
}

void textview_destroy(TextView *textview)
{
	textview_uri_list_remove_all(textview->uri_list);
	textview->uri_list = NULL;
	g_free(textview);
}

void textview_quote_color_set(void)
{
}

void textview_set_font(void)
{
	static gint prev_ascent = 0, prev_descent = 0;

	if (prefs_common.messagefont) {
		if (msgfont) {
			msgfont->ascent = prev_ascent;
			msgfont->descent = prev_descent;
			gdk_font_unref(msgfont);
		}
		msgfont = gdk_fontset_load(prefs_common.messagefont);
		if (msgfont) {
			gint ascent, descent;

			prev_ascent = msgfont->ascent;
			prev_descent = msgfont->descent;
			descent = prefs_common.line_space / 2;
			ascent  = prefs_common.line_space - descent;
			msgfont->ascent  += ascent;
			msgfont->descent += descent;
		}
	}
	if (!boldfont)
		boldfont = gdk_fontset_load(BOLD_FONT);
	if (!spacingfont)
		spacingfont = gdk_fontset_load(SPACING_FONT);
}

enum
{
	H_DATE		= 0,
	H_FROM		= 1,
	H_TO		= 2,
	H_NEWSGROUPS	= 3,
	H_SUBJECT	= 4,
	H_CC		= 5,
	H_REPLY_TO	= 6,
	H_FOLLOWUP_TO	= 7,
	H_X_MAILER	= 8,
	H_X_NEWSREADER	= 9,
	H_USER_AGENT	= 10,
	H_ORGANIZATION	= 11,
};

static void textview_scan_header(TextView *textview, FILE *fp, const gchar *charset)
{
	static HeaderEntry hentry[] = {{"Date:",         NULL, FALSE},
				       {"From:",         NULL, TRUE},
				       {"To:",	         NULL, FALSE},
				       {"Newsgroups:",   NULL, FALSE},
				       {"Subject:",      NULL, TRUE},
				       {"Cc:",	         NULL, FALSE},
				       {"Reply-To:",     NULL, FALSE},
				       {"Followup-To:",  NULL, FALSE},
				       {"X-Mailer:",     NULL, TRUE},
				       {"X-Newsreader:", NULL, TRUE},
				       {"User-Agent:",   NULL, TRUE},
				       {"Organization:", NULL, TRUE},
				       {"S:",		 NULL, TRUE},
				       {"SSV:",		 NULL, TRUE},
				       {"R:",		 NULL, TRUE},
				       {NULL,	         NULL, FALSE}};
	gchar buf[BUFFSIZE], tmp[BUFFSIZE];
	GtkText *text = GTK_TEXT(textview->text);
	gint hnum;
	HeaderEntry *hp;

	g_return_if_fail(fp != NULL);

	if (!prefs_common.display_header) {
		while (fgets(buf, sizeof(buf), fp) != NULL)
			if (buf[0] == '\r' || buf[0] == '\n') break;
		return;
	}

	gtk_text_freeze(text);

	/* show specified headers */
	while ((hnum = procheader_get_one_field(buf, sizeof(buf), fp, hentry))
	       != -1) {
		hp = hentry + hnum;

		gtk_text_insert(text, boldfont, NULL, NULL, buf,
				strlen(hp->name));
		conv_unmime_header(tmp, sizeof(tmp), buf + strlen(hp->name),
				   charset);
		if (hnum == H_X_MAILER && prefs_common.enable_color &&
		    strstr(tmp, "stuphead") != NULL)
			gtk_text_insert(text, normalfont, &prefs_common.emphasis_color,
					NULL, tmp, -1);
		else
			gtk_text_insert(text, normalfont, NULL, NULL, tmp, -1);
		gtk_text_insert(text, NULL, NULL, NULL, "\n", 1);
	}

	gtk_text_insert(text, NULL, NULL, NULL, "\n", 1);
	gtk_text_thaw(text);
}

void textview_scroll_one_line(TextView *textview, gboolean up)
{
	GtkText *text = GTK_TEXT(textview->text);
	gfloat upper;

	if (!up) {
		upper = text->vadj->upper - text->vadj->page_size;
		if (text->vadj->value < upper) {
			text->vadj->value +=
				text->vadj->step_increment * 4;
			text->vadj->value =
				MIN(text->vadj->value, upper);
			gtk_signal_emit_by_name(GTK_OBJECT(text->vadj),
						"value_changed");
		}
	} else {
		if (text->vadj->value > 0.0) {
			text->vadj->value -=
				text->vadj->step_increment * 4;
			text->vadj->value =
				MAX(text->vadj->value, 0.0);
			gtk_signal_emit_by_name(GTK_OBJECT(text->vadj),
						"value_changed");
		}
	}
}

gboolean textview_scroll_page(TextView *textview, gboolean up)
{
	GtkText *text = GTK_TEXT(textview->text);
	gfloat upper;

	if (!up) {
		upper = text->vadj->upper - text->vadj->page_size;
		if (text->vadj->value < upper) {
			text->vadj->value += text->vadj->page_increment;
			text->vadj->value = MIN(text->vadj->value, upper);
			gtk_signal_emit_by_name(GTK_OBJECT(text->vadj),
						"value_changed");
		} else
			return FALSE;
	} else {
		if (text->vadj->value > 0.0) {
			text->vadj->value -= text->vadj->page_increment;
			text->vadj->value = MAX(text->vadj->value, 0.0);
			gtk_signal_emit_by_name(GTK_OBJECT(text->vadj),
						"value_changed");
		} else
			return FALSE;
	}

	return TRUE;
}

static void textview_key_pressed(GtkWidget *widget, GdkEventKey *event,
				 TextView *textview)
{
	SummaryView *summaryview = NULL;

	if (!event) return;
	if (textview->messageview->mainwin)
		summaryview = textview->messageview->mainwin->summaryview;

	switch (event->keyval) {
	case GDK_Tab:
	case GDK_Home:
	case GDK_Left:
	case GDK_Up:
	case GDK_Right:
	case GDK_Down:
	case GDK_Page_Up:
	case GDK_Page_Down:
	case GDK_End:
	case GDK_Control_L:
	case GDK_Control_R:
		return;
	case GDK_space:
		if (summaryview)
			summary_pass_key_press_event(summaryview, event);
		else
			textview_scroll_page(textview, FALSE);
		break;
	case GDK_BackSpace:
		textview_scroll_page(textview, TRUE);
		break;
	case GDK_KP_Enter:
	case GDK_Return:
		textview_scroll_one_line(textview,
					 (event->state & GDK_MOD1_MASK) != 0);
		break;
	default:
		if (summaryview)
			summary_pass_key_press_event(summaryview, event);
		break;
	}

	if (summaryview)
		gtk_widget_grab_focus(summaryview->ctree);
}

static void textview_button_pressed(GtkWidget *widget, GdkEventButton *event,
				    TextView *textview)
{
	if (event &&
	    ((event->button == 1 && event->type == GDK_2BUTTON_PRESS) ||
	     event->button == 2)) {
		GSList *cur;
		guint current_pos = GTK_EDITABLE(textview->text)->current_pos;

		for (cur = textview->uri_list; cur != NULL; cur = cur->next) {
			RemoteURI *uri = (RemoteURI *)cur->data;

			if (current_pos >= uri->start &&
			    current_pos < uri->end)
				textview_open_uri(uri->uri);
		}
	}
}

static void textview_uri_list_remove_all(GSList *uri_list)
{
	GSList *cur;

	for (cur = uri_list; cur != NULL; cur = cur->next) {
		if (cur->data) {
			g_free(((RemoteURI *)cur->data)->uri);
			g_free(cur->data);
		}
	}

	g_slist_free(uri_list);
}

void textview_open_uri(const gchar *uri)
{
	static gchar *default_cmdline = "netscape -remote openURL(%s,raise)";
	gchar buf[1024];
	gchar *p;

	if (prefs_common.uri_cmd &&
	    (p = strchr(prefs_common.uri_cmd, '%')) && *(p + 1) == 's' &&
	    !strchr(p + 2, '%'))
		g_snprintf(buf, sizeof(buf), prefs_common.uri_cmd, uri);
	else {
		if (prefs_common.uri_cmd)
			g_warning(_("Open URI command line is invalid: `%s'"),
				  prefs_common.uri_cmd);
		g_snprintf(buf, sizeof(buf), default_cmdline, uri);
	}

	execute_command_line(buf);
};
